"""
Functions for compatibility with MyGrad to allow automatic differentiation. Includes:
    get_matrix_function() which returns a mygrad compatible version of the _matrix_function() in myokit markov models.
    linalg_solve() which is a mygrad compatible matrix Ax=b equation solver.
"""
import myokit
import myokit.lib.markov
import numpy as np
import mygrad as mg


# =============================================================================
#
# =============================================================================


#  Code taken from https://github.com/myokit/myokit/blob/main/myokit/lib/markov.py on 2024/02/07 (last edit was commit https://github.com/myokit/myokit/commit/31fe43cea4af9efcfb0d428e056e5d82173170c3) under the following license. It has been changed to work with MyGrad rather than numpy and with comments to ignore some code for coverage calculations.
# =============================================================================
# BSD 3-Clause License
#
# Copyright (c) 2011-2017 Maastricht University. All rights reserved.
# Copyright (c) 2017-2020 University of Oxford. All rights reserved.
# (University of Oxford means the Chancellor, Masters and Scholars of the University of #Oxford, having an administrative office at Wellington Square, Oxford OX1 2JD, UK).
# Copyright (c) 2020-2024 University of Nottingham. All rights reserved.
# =============================================================================

# noinspection PyProtectedMember,PyListCreation,PyUnboundLocalVariable,PyUnusedLocal,GrazieInspection
def get_matrix_function(model):
    """
    Generates a MyGrad compatible _matrix_function for the automatic differentiation in the calculation of Markov model steady states. 
    Contains edits to swap out the numpy transition matrix with a MyGrad tensor.
    Removed any code that edits the model.
    Some edits to allow the code to work outside myokit.lib.markov
    Unnecessary code for the matrix function has been commented out.
    """
    # Create mapping from states to index
    state_indices = {}
    for k, state in enumerate(model._states):
        state_indices[state] = k

    # Get expressions for state & current variables, but with all
    # references to variables (except states and parameters) replaced by
    # inlined expressions.
    expressions = []
    for v in model._states:
        expressions.append(v.rhs().clone(expand=True, retain=model._inputs))
    if model._current is not None:
        current_expression = model._current.rhs().clone(
            expand=True, retain=model._inputs)

    # Create parametrisable matrices to evaluate the state & current
    # This checks that each state's RHS can be written as a linear
    # combination of the states, and gathers the corresponding multipliers.

    # Matrix of linear factors
    n = len(model._states)
    A = [[myokit.Number(0) for j in range(n)] for i in range(n)]

    # # List of transitions
    # T = set()

    # Populate A and T
    for row, e in enumerate(expressions):

        # Check if this expression is a linear combination of the states
        try:
            factors = myokit.lib.markov._linear_combination(e, model._states)
        except ValueError:  # pragma: no cover
            raise myokit.lib.markov.LinearModelError(
                'Unable to write expression as linear combination of'
                ' states: ' + str(e) + '.')

        # Scan factors
        for col, state in enumerate(model._states):
            factor = factors[col]
            if factor is not None:
                # Add factor to transition matrix
                A[row][col] = factor

                # # Store transition in transition list
                # if row != col:
                #     T.add((col, row))   # A is mirrored

    # Create a parametrisable matrix for the current
    B = [myokit.Number(0) for i in range(n)]
    if model._current is not None:
        try:
            factors = myokit.lib.markov._linear_combination(current_expression, model._states)
        except ValueError:  # pragma: no cover
            raise myokit.lib.markov.LinearModelError(
                'Unable to write expression as linear combination of'
                ' states: ' + str(e) + '.')

        for col, state in enumerate(model._states):
            factor = factors[col]
            if factor is not None:
                B[col] = factor

    # # Create list of transition rates and associated equations
    # T = list(T)
    # T.sort()
    # R = []
    # for i in range(len(A)):
    #     for j in range(len(A)):
    #         if (i, j) in T:
    #             R.append((i, j, A[j][i]))   # A is mirrored
    # del T

    #
    # Create function to create parametrisable matrices
    #
    model._model.reserve_unique_names('A', 'B', 'n', 'numpy')
    writer = myokit.numpy_writer()
    w = writer.ex
    head = 'def matrix_function('
    head += ','.join([w(p.lhs()) for p in model._inputs])
    head += '):'
    body = []
    body.append('A = mg.zeros((n, n), dtype=numpy.double)')
    zero = myokit.Number(0)
    for i, row in enumerate(A):
        for j, e in enumerate(row):
            if e != zero:
                body.append('A[' + str(i) + ',' + str(j) + '] = ' + w(e))
    body.append('B = mg.zeros(n, dtype=numpy.double)')
    for j, e in enumerate(B):
        if e != zero:
            body.append('B[' + str(j) + '] = ' + w(e))
    body.append('return A, B')
    code = head + '\n' + '\n'.join(['    ' + line for line in body])
    globl = {'numpy': np, 'n': n, 'mg': mg}
    local = {}

    exec(code, globl, local)
    # model._matrix_function = local['matrix_function']

    # #
    # # Create function to return list of transition rates
    # #
    # model._model.reserve_unique_names('R', 'n', 'numpy')
    # head = 'def rate_list_function('
    # head += ','.join([w(p.lhs()) for p in model._inputs])
    # head += '):'
    # body = []
    # body.append('R = []')
    # for i, j, e in R:
    #     body.append(
    #         'R.append((' + str(i) + ',' + str(j) + ',' + w(e) + '))')
    # body.append('return R')
    # code = head + '\n' + '\n'.join(['    ' + line for line in body])
    # globl = {'numpy': np}
    # local = {}
    # exec(code, globl, local)
    # model._rate_list_function = local['rate_list_function']

    return local['matrix_function']


def linalg_solve(A, B):
    """
    Solve the linear system Ax=B for x. Uses MyGrad tensors to ensure automatic differentiation works.
    Performs Gaussian elimination with partial pivoting followed by back substitution.

    Parameters
    ----------
    A : MyGrad tensor
        Square matrix A.
    B : MyGrad tensor
        A vector of length matching the size of A.

    Returns
    -------
    x : MyGrad tensor
        The solution to Ax=B with all derivative information included.
    """
    x = mg.zeros(B.shape, dtype=np.double)
    n = B.size
    pivot = list(range(n))
    for row in range(n - 1):
        # Partial pivoting
        for i in range(row + 1, n):
            if np.abs(A[pivot[row], row]) < np.abs(A[pivot[i], row]):
                pivot[row], pivot[i] = pivot[i], pivot[row]
        # Gaussian elimination
        for i in range(row + 1, n):
            ratio = A[pivot[i], row] / A[pivot[row], row]
            A[pivot[i], :] -= ratio * A[pivot[row], :]
            B[pivot[i]] -= ratio * B[pivot[row]]
    # Zero out any rounding errors in A
    for row in range(n):
        for i in range(row):
            A[pivot[row], i] = 0
    # Back substitution
    x[n - 1] = B[pivot[n - 1]] / A[pivot[n - 1], n - 1]

    for row in range(n - 2, -1, -1):
        summ = A[pivot[row], :] @ x
        x[row] = (B[pivot[row]] - summ) / A[pivot[row], row]
    return x



import ionbench.problems.staircase
import scipy.optimize
from ionbench.utils.scipy_setup import minimize_bounds


# noinspection PyShadowingNames
def run(bm, x0=None, maxIter=1000, debug=False):
    """
    Runs Sequential Least SQuares Programming optimiser from Scipy. An example of a Sequential Quadratic Programming method which uses a quasi-newton update strategy to approximate the hessian.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter vector from which to start optimisation. Default is None, in which case a randomly sampled parameter vector is retrieved from bm.sample().
    maxIter : int, optional
        Maximum number of iterations of SLSQP to use. The default is 1000.
    debug : bool, optional
        If True, prints out the cost and parameters found by the algorithm. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified by SLSQP.
    """
    if x0 is None:
        x0 = bm.sample()
        if debug:
            print('Sampling x0')
            print(x0)
    if bm.ratesBounded:
        constraints = {'type': 'eq', 'fun': lambda p: bm.parameter_penalty(bm.original_parameter_space(p))+bm.rate_penalty(bm.original_parameter_space(p))}
    else:
        constraints = ()
    if bm.parametersBounded:
        bounds = minimize_bounds(bm)
        out = scipy.optimize.minimize(bm.cost, x0, jac=bm.grad, method='SLSQP', constraints=constraints, options={'disp': debug, 'maxiter': maxIter}, bounds=bounds)
    else:
        out = scipy.optimize.minimize(bm.cost, x0, jac=bm.grad, method='SLSQP', constraints=constraints, options={'disp': debug, 'maxiter': maxIter})

    if out.nit >= maxIter:
        bm.set_max_iter_flag()

    if debug:
        print(f'Cost of {out.fun} found at:')
        print(out.x)

    bm.evaluate()
    return out.x


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> BuenoOrovio2008

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so BuenoOrovio2008.

    """
    mod = ionbench.modification.BuenoOrovio2008()
    return mod



from scipy.stats import beta
import numpy as np


def bootstrap_success_rate(success):
    """
    Draw a smooth bootstrapped sample of the success rate based on the observed successes.
    Parameters
    ----------
    success : list
        A list of booleans indicating whether each run was successful.

    Returns
    -------
    successRate : float
        The success rate of the sample.
    """

    return beta.rvs(np.sum(success) + 0.5, len(success) - np.sum(success) + 0.5)


def bootstrap_ERT(successes, times):
    """
    Generate a bootstrap sample of ERT.
    Parameters
    ----------
    successes : list
        A list of booleans indicating whether each run was successful.
    times : np.array
        Vector of times to generate the bootstrap sample from.
    Returns
    -------
    u : np.array
        The bootstrap sample of ERT.
    """
    count = len(successes)
    mask = np.random.choice(count, count)
    successes = successes[mask]
    times = times[mask]
    ERT = expected_time(times, successes, bootstrap=True)
    return ERT


def expected_time(times, successes, bootstrap=False):
    """
    Calculate the expected time to solve a problem using the median-unbiased estimator of the success rate.
    Parameters
    ----------
    times : list
        A list of times to solve the problem.
    successes : list
        A list of booleans indicating whether each run was successful.
    bootstrap : bool, optional
        If True, then whenever successes contains only True or only False, the corresponding missing time information is taken as a random sample from the times. Otherwise, the average of all times is taken.

    Returns
    -------
    expectedTime : float
        The expected time to successfully solve the problem.
    """
    if np.all(successes) or not np.any(successes):
        if np.all(successes) and bootstrap:
            Tfail = np.random.choice(times)
        else:
            Tfail = np.mean(times)
        if not np.any(successes) and bootstrap:
            Tsucc = np.random.choice(times)
        else:
            Tsucc = np.mean(times)
    else:
        Tsucc = np.mean([times[i] for i in range(len(times)) if successes[i]])
        Tfail = np.mean([times[i] for i in range(len(times)) if not successes[i]])
    if bootstrap:
        success_rate = bootstrap_success_rate(successes)
    else:
        success_rate = np.mean(successes)
    if success_rate == 0:
        return np.inf
    expectedTime = Tsucc + Tfail * (1 - success_rate) / success_rate
    return expectedTime



"""
The module provides the PSO algorithm given by Loewe et al. 2016.
The particles are clamped differently to other PSO algorithms.
No initial position or velocity sampling given, so we assume ionBench defaults.
"""
import numpy as np
import ionbench
import warnings


# noinspection PyShadowingNames
def run(bm, x0=None, n=96, maxIter=1000, phi1=2.05, phi2=2.05, debug=False):
    """
    Runs the particle swarm optimisation from Loewe et al. 2016. If the benchmarker is bounded, the solver will search in the interval [lb,ub], otherwise the solver will search in the interval [0,2*default].

    Notes on using this optimiser:
        It is unclear the specifics of the optimisation used in Loewe et al. 2016. It does not describe the method for sampling the initial positions or velocities of the particles (we use uniform distribution between bounds for position and expect Loewe et al. 2016 did similarly, and zeros for the initial velocities, waiting for them to be set by TRR). The initial velocities may be different to those used in the Loewe et al. 2016.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter guess. Population is generated by randomly perturbing this initial guess +-50%, then applying the appropriate bounds. If x0=None (the default), then the population will be sampled uniformly between the bounds.
    n : int, optional
        Number of particles. The default is 96.
    maxIter : int, optional
        Maximum number of iterations. The default is 1000.
    phi1 : float, optional
        Scale of the acceleration towards a particle's best positions. The default is 2.05.
    phi2 : float, optional
        Scale of the acceleration towards the best point seen across all particles. The default is 2.05.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """
    if not bm.parametersBounded:
        raise RuntimeError('This optimiser requires bounds.')

    if x0 is None:
        x0 = bm.sample()

    signed_error = ionbench.utils.cache.get_cached_signed_error(bm)

    def cost_func(x):
        """
        Cost function needs to use the signed_error function so that they use the same cache.
        """
        return bm.rmse(signed_error(x) + bm.DATA, bm.DATA)

    # noinspection PyShadowingNames
    class Particle(ionbench.utils.particle_optimisers.Particle):
        def __init__(self):
            super().__init__(bm, cost_func, x0)

        def clamp(self):
            """
            Clamp parameters to the input parameter space.
            """
            for i in range(bm.n_parameters()):
                if self.position[i] < 0:  # pragma: no cover
                    self.position[i] = np.random.rand() / 4
                elif self.position[i] > 1:  # pragma: no cover
                    self.position[i] = np.random.rand() / 4

    L = None

    if (phi1 + phi2) ** 2 - 4 * (phi1 + phi2) < 0:
        warnings.warn(
            "Invalid constriction factor using specified values for phi1 and phi2. Using defaults of phi1=phi2=2.05 instead.")
        phi1 = 2.05
        phi2 = 2.05
    phi = phi1 + phi2
    constFactor = 2 / (phi - 2 + np.sqrt(phi ** 2 - 4 * phi))

    particleList = []
    for i in range(n):
        particleList.append(Particle())

    Gcost = [np.inf] * maxIter  # Best cost ever
    Gpos = [None] * maxIter  # Position of best cost ever
    for L in range(maxIter):
        if L > 0:
            Gcost[L] = Gcost[L - 1]
            Gpos[L] = Gpos[L - 1]

        if debug:
            print('-------------')
            print(f'Beginning population: {L}')
            print(f'Best cost so far: {Gcost[L]}')
            print(f'Found at position: {Gpos[L]}')

        # Find best positions, both globally and locally
        for p in particleList:
            p.set_cost()
            if p.currentCost < Gcost[L]:
                Gcost[L] = p.currentCost
                Gpos[L] = np.copy(p.position)

        # Update velocities
        for p in particleList:
            # Update velocity
            localAcc = phi1 * np.random.rand() * (p.bestPosition - p.position)
            globalAcc = phi2 * np.random.rand() * (Gpos[L] - p.position)
            p.velocity = constFactor * (p.velocity + localAcc + globalAcc)

            # Move particle
            p.position += p.velocity

            # Clamp
            p.clamp()

        if debug:
            print("Positions renewed")
            print(f'Finished population {L}')
            print(f'Best cost so far: {Gcost[L]}')
            print(f'Found at position: {Gpos[L]}')

        if bm.is_converged():
            break

    if L >= maxIter-1:
        bm.set_max_iter_flag()

    bm.evaluate()
    return Particle().untransform(Gpos[L])


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Loewe2016

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Loewe2016.

    """
    mod = ionbench.modification.Loewe2016()
    return mod



import numpy as np


class Particle:
    """
    Default particle class for ionbench optimisers. Optimisers may use subclasses which override the set_position and set_velocity methods.
    Particles operate on the interval [0,1] which is then mapped to [lb,ub] for all optimisers to evaluate the cost.
    """
    def __init__(self, bm, cost_func, x0):
        if bm.parametersBounded:
            self.lb = bm.input_parameter_space(bm.lb)
            self.ub = bm.input_parameter_space(bm.ub)
        else:
            self.lb = None
            self.ub = None
        self.cost_func = cost_func
        self.position = np.zeros(len(x0))
        self.velocity = np.zeros(len(x0))
        self.currentCost = None
        self.set_position(x0)
        self.set_velocity()
        self.bestCost = np.inf  # Best cost of this particle
        self.bestPosition = np.copy(self.position)  # Position of best cost for this particle

    def set_position(self, x0):
        self.position = self.transform(x0 * np.random.uniform(low=0.9, high=1.1, size=len(x0)))
        self.clamp()

    def set_velocity(self):
        self.velocity = 0.1 * np.random.rand(len(self.position))

    def set_cost(self, cost=None):
        if cost is None:
            self.currentCost = self.cost_func(self.untransform(self.position))
        else:
            self.currentCost = cost
        if self.currentCost < self.bestCost:
            self.bestCost = self.currentCost
            self.bestPosition = np.copy(self.position)

    def transform(self, x):
        """
        Map from input space to [0,1] bounded space.

        Parameters
        ----------
        x : np.ndarray
            The parameters to transform.

        Returns
        -------
        xTrans : np.ndarray
            The transformed parameters.
        """
        xTrans = (x-self.lb) / (self.ub - self.lb)
        return xTrans

    def untransform(self, parameters):
        """
        Map from [0,1] to [lb,ub] in input parameter space.

        Parameters
        ----------
        parameters : np.ndarray
            The parameters to untransform.

        Returns
        -------
        xTrans : np.ndarray
            The untransformed parameters.
        """
        xTrans = parameters * (self.ub - self.lb) + self.lb
        return xTrans

    def clamp(self):
        """
        Clamp parameters to the input parameter space.
        """
        self.position = np.clip(self.position, 0, 1)



"""
Contains the two staircase benchmarker problems, HH and MM. These both inherit from StaircaseBenchmarker which itself inherits from benchmarker.Benchmarker.
generate_data(modelType) will generate the data for either the HH or MM problem and store it in the data directory.
"""
import csv
import os

import myokit
import numpy as np
from scipy.stats import loguniform
from myokit.lib.hh import HHModel
from myokit.lib.markov import LinearModel

import ionbench


# noinspection PyUnresolvedReferences
class StaircaseBenchmarker(ionbench.benchmarker.Benchmarker):
    def __init__(self):
        # Benchmarker
        p = self.protocol()
        self.TIMESTEP = 0.5  # Timestep in data between points
        self.T_MAX = p.characteristic_time()
        self._LOWER_BOUND = np.array([1e-7] * (self.n_parameters() - 1) + [0.02])
        self._UPPER_BOUND = np.array(
            [1e3 if self.STANDARD_LOG_TRANSFORM[i] else 0.4 for i in range(self.n_parameters() - 1)] + [0.2])
        self.lb = np.copy(self._LOWER_BOUND)
        self.ub = np.copy(self._UPPER_BOUND)
        self.RATE_MIN = 1.67e-5
        self.RATE_MAX = 1e3
        try:
            self.load_data(
                os.path.join(ionbench.DATA_DIR, 'staircase', 'data' + ('HH' if 'hh' in self.NAME else 'MM') + '.csv'))
        except FileNotFoundError:  # pragma: no cover
            self.DATA = None

        # Myokit
        self.sim = myokit.Simulation(self._MODEL, protocol=p)
        self.sim.set_tolerance(*self._TOLERANCES)
        if self.sensitivityCalc:
            paramNames = [self._PARAMETER_CONTAINER + '.p' + str(i + 1) for i in range(self.n_parameters())]
            sens = ([self._OUTPUT_NAME], paramNames)
            self.simSens = myokit.Simulation(self._MODEL, sensitivities=sens, protocol=p)
            self.simSens.set_tolerance(*self._TOLERANCES)
        else:
            self.simSens = None
        super().__init__()

    def sample(self, n=1):
        """
        Sample parameters for the staircase problems.

        Parameters
        ----------
        n : int, optional
            Number of parameter vectors to sample. The default is 1.

        Returns
        -------
        params : list
            If n=1, then params is the vector of parameters. Otherwise, params is a list containing n parameter vectors.

        """
        params = [None] * n
        for i in range(n):
            while True:
                p = []
                for j in range(self.n_parameters()):
                    if self.STANDARD_LOG_TRANSFORM[j]:
                        p.append(loguniform.rvs(self._LOWER_BOUND[j], self._UPPER_BOUND[j]))
                    else:
                        p.append(np.random.uniform(self._LOWER_BOUND[j], self._UPPER_BOUND[j]))
                if self.in_rate_bounds(p, boundedCheck=False):
                    break
            params[i] = self.input_parameter_space(p)
        if n == 1:
            return params[0]
        else:
            return params

    @staticmethod
    def protocol():
        """
        Gets the staircase voltage protocol from the loaded log and returns it. Setting self.T_MAX to the length of the protocol.

        Returns
        -------
        protocol : myokit.Protocol
            The staircase protocol.
        """
        protocol = myokit.load_protocol(os.path.join(ionbench.DATA_DIR, 'staircase', 'staircase-pace.mmt'))
        return protocol

    @staticmethod
    def add_ramps(model):
        """
        Myokit protocols do not support ramps, so this method is used to add the staircase ramps to the myokit model. This does not copy the model and will modify the input.

        Parameters
        ----------
        model : myokit.Model
            The model to add the ramps to.

        Returns
        -------
        model : myokit.Model
            The model with the ramps added.
        """
        c = model.get('membrane')
        # Remove binding from membrane.V
        v = c.get('V')
        v.set_binding(None)

        # Ramps
        v1 = c.add_variable('v1')
        v1.set_rhs('-150 + 0.1 * engine.time')
        v2 = c.add_variable('v2')
        v2.set_rhs('5694 - 0.4 * engine.time')

        # membrane.V is pace with ramps
        v.set_rhs("""
            piecewise(
                (engine.time >= 300 and engine.time < 700), v1, 
                (engine.time >=14410 and engine.time < 14510), v2, 
                engine.pace)
        """)
        return model


class HH(StaircaseBenchmarker):
    """
    The Hodgkin-Huxley IKr Staircase benchmarker.

    The benchmarker uses the Beattie et al. 2017 IKr Hodgkin-Huxley model using the staircase protocol.

    Its parameters are specified as scaling factors, so start at a vector of all ones, or sample from the benchmarker.sample() method.
    """

    def __init__(self, sensitivities=False):
        print('Initialising Hodgkin-Huxley IKr benchmark')
        # Benchmarker
        self.NAME = "staircase.hh"
        self._TRUE_PARAMETERS = np.array([2.26e-4, 0.0699, 3.45e-5, 0.05462, 0.0873, 8.91e-3, 5.15e-3, 0.03158, 0.1524])
        self.STANDARD_LOG_TRANSFORM = (True, False) * 4 + (False,)
        self._RATE_FUNCTIONS = (lambda p, V: p[0] * np.exp(p[1] * V),
                                lambda p, V: p[2] * np.exp(-p[3] * V),
                                lambda p, V: p[4] * np.exp(p[5] * V),
                                lambda p, V: p[6] * np.exp(-p[7] * V))  # Used for rate bounds
        self.sensitivityCalc = sensitivities
        self.COST_THRESHOLD = 1.57e-2

        # Myokit
        model = myokit.load_model(os.path.join(ionbench.DATA_DIR, 'staircase', 'beattie-2017-ikr-hh.mmt'))
        self._MODEL = self.add_ramps(model)
        self._OUTPUT_NAME = 'ikr.IKr'
        self._PARAMETER_CONTAINER = 'ikr'
        self._ANALYTICAL_MODEL = HHModel(model=self._MODEL, states=['ikr.act', 'ikr.rec'], current=self._OUTPUT_NAME,
                                         parameters=[self._PARAMETER_CONTAINER + '.p' + str(i + 1) for i in
                                                     range(self.n_parameters())], vm='membrane.V')
        self._TOLERANCES = (1e-5, 1e-5)

        super().__init__()
        print('Benchmarker initialised')


class MM(StaircaseBenchmarker):
    """
    The Markov IKr Staircase benchmarker.

    The benchmarker uses the Fink et al. 2008 IKr Markov model using the staircase protocol.

    Its parameters are specified as scaling factors, so start at a vector of all ones, or sample from the benchmarker.sample() method.
    """

    def __init__(self, sensitivities=False):
        print('Initialising Markov Model IKr benchmark')
        # Benchmarker
        self.NAME = "staircase.mm"
        self._TRUE_PARAMETERS = np.array(
            [0.20618, 0.0112, 0.04209, 0.02202, 0.0365, 0.41811, 0.0223, 0.13279, 0.0603, 0.08094, 0.0002262, 0.0399,
             0.04150, 0.0312, 0.024])
        self.STANDARD_LOG_TRANSFORM = (True, False, True) * 2 + (False, True) * 2 + (True, False) * 2 + (False,)
        self._RATE_FUNCTIONS = (lambda p, V: p[0] * np.exp(p[1] * V),
                                lambda p, V: p[2],
                                lambda p, V: p[3] * np.exp(p[4] * V),
                                lambda p, V: p[5] * np.exp(p[6] * V),
                                lambda p, V: p[7] * np.exp(-p[8] * V),
                                lambda p, V: p[9],
                                lambda p, V: p[10] * np.exp(-p[11] * V),
                                lambda p, V: p[12] * np.exp(-p[13] * V))  # Used for rate bounds
        self.sensitivityCalc = sensitivities
        self.COST_THRESHOLD = 5.77e-3

        # Myokit
        model = myokit.load_model(os.path.join(ionbench.DATA_DIR, 'staircase', 'fink-2008-ikr-mm.mmt'))
        self._MODEL = self.add_ramps(model)
        self._OUTPUT_NAME = 'IKr.i_Kr'
        self._PARAMETER_CONTAINER = 'iKr_Markov'
        self._ANALYTICAL_MODEL = LinearModel(model=self._MODEL, current=self._OUTPUT_NAME, vm='membrane.V',
                                             states=['iKr_Markov.' + s for s in ['Cr1', 'Cr2', 'Cr3', 'Or4', 'Ir5']],
                                             parameters=[self._PARAMETER_CONTAINER + '.p' + str(i + 1) for i in
                                                         range(self.n_parameters())])
        self._TOLERANCES = (1e-6, 1e-6)
        super().__init__()
        print('Benchmarker initialised')


# noinspection PyProtectedMember
def generate_data(modelType, noiseStrength=0.05):  # pragma: no cover
    """
    Generate the data files for the staircase benchmarker problems.

    Parameters
    ----------
    modelType : string
        'HH' to generate the data for the Hodgkin-Huxley benchmark problem. 'MM' to generate the data for the Markov model benchmark problem.
    noiseStrength : float, optional
        The strength of the noise to add to the data. The default is 0.05, which is 5% of the mean absolute value of the data.

    Returns
    -------
    None.

    """
    modelType = modelType.upper()
    if modelType == 'HH':
        bm = HH()
    else:
        bm = MM()
    bm.set_params(bm._TRUE_PARAMETERS)
    bm.set_steady_state(bm._TRUE_PARAMETERS)
    out = bm.solve_model(np.arange(0, bm.T_MAX, bm.TIMESTEP), continueOnError=False)
    out += np.random.normal(0, np.mean(np.abs(out)) * noiseStrength, len(out))
    with open(os.path.join(ionbench.DATA_DIR, 'staircase', 'data' + modelType + '.csv'), 'w', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',')
        writer.writerows(map(lambda x: [x], out))



import pytest
import ionbench
import numpy as np
from importlib import import_module
from inspect import signature


class TestScipy:
    # Check all scipy optimisers run and improve cost compared to x0
    bmTest = ionbench.problems.test.Test()

    @pytest.mark.parametrize("opt", ionbench.OPT_SCIPY)
    @pytest.mark.filterwarnings("ignore::RuntimeWarning")
    @pytest.mark.cheap
    def test_minimum_finding(self, opt):
        self.bmTest.reset()
        module = import_module(opt)
        mod = module.get_modification()
        mod.apply(self.bmTest)
        x0_opt = module.run(self.bmTest, debug=True)
        cost_opt = self.bmTest.cost(x0_opt)
        assert cost_opt < 5e-3


class TestPints:
    bmTest = ionbench.problems.test.Test()

    @pytest.mark.parametrize("opt", ionbench.OPT_PINTS)
    @pytest.mark.filterwarnings("ignore::RuntimeWarning")
    @pytest.mark.cheap
    def test_minimum_finding(self, opt):
        self.bmTest.reset()
        module = import_module(opt)
        mod = module.get_modification()
        mod.apply(self.bmTest)
        if 'cmaes' in opt:
            # Turn off bounds since rate bounds don't make sense here
            self.bmTest._parameters_bounded = False
        x0_opt = module.run(self.bmTest, debug=True)
        cost_opt = self.bmTest.cost(x0_opt)
        assert cost_opt < 5e-3


class TestExternal:
    bmTest = ionbench.problems.test.Test()

    @pytest.mark.parametrize("opt", ionbench.OPT_EXT)
    @pytest.mark.filterwarnings("ignore::RuntimeWarning")
    @pytest.mark.cheap
    def test_minimum_finding(self, opt):
        self.bmTest.reset()
        module = import_module(opt)
        mod = module.get_modification()
        mod.apply(self.bmTest)
        if 'maxIter' in signature(module.run).parameters:
            if 'loewe' in opt.lower():
                x0_opt = module.run(self.bmTest, n=5, maxIter=10000, debug=True)
            else:
                x0_opt = module.run(self.bmTest, maxIter=10000, debug=True)
        else:
            x0_opt = module.run(self.bmTest, debug=True)
        cost_opt = self.bmTest.cost(x0_opt)
        assert cost_opt < 5e-3


class TestMaxIterFlag:
    bmTest = ionbench.problems.test.Test()
    bmTest.COST_THRESHOLD = 0
    @pytest.mark.parametrize("opt", ionbench.OPT_ALL)
    @pytest.mark.filterwarnings("ignore::RuntimeWarning")
    @pytest.mark.cheap
    def test_max_iter_flag(self, opt):
        self.bmTest.reset()
        module = import_module(opt)
        mod = module.get_modification()
        mod.apply(self.bmTest)
        inputs = {}
        if 'maxIter' in signature(module.run).parameters:
            inputs['maxIter'] = 2
        if 'nGens' in signature(module.run).parameters:
            inputs['nGens'] = 2
        if 'n' in signature(module.run).parameters:
            inputs['n'] = 5
        if 'popSize' in signature(module.run).parameters:
            inputs['popSize'] = 20
        module.run(self.bmTest, debug=True, **inputs)
        assert self.bmTest.tracker.maxIterFlag is True



"""
Contains the main Benchmarker class from which the benchmarker problems inherit the vast majority of their methods.
Contains the Tracker class which is stored as an attribute in each benchmarker and managers tracking the performance metrics over an optimisation.
"""
import numpy as np
import myokit
import myokit.lib.markov
import csv
import ionbench
import matplotlib.pyplot as plt
import warnings
import time
import mygrad as mg

from ionbench.tracker import Tracker


class Benchmarker:
    """
    The Benchmarker class contains all the features needed to evaluate an optimisation algorithm. This class should not need to be called directly and is instead used as a parent class for the benchmarker problems.

    The main methods to use from this class are n_parameters(), cost(), reset(), and evaluate().
    """

    def __init__(self):
        # Ensure any undefined attributes (those defined by subclasses) are defined if missing
        if not hasattr(self, 'DATA'):  # pragma: no cover
            self.DATA = np.array(None)
        if not hasattr(self, '_LOWER_BOUND'):  # pragma: no cover
            self._LOWER_BOUND = np.array(None)
        if not hasattr(self, '_UPPER_BOUND'):  # pragma: no cover
            self._UPPER_BOUND = np.array(None)
        if not hasattr(self, 'lb'):  # pragma: no cover
            self.lb = np.array(None)
        if not hasattr(self, 'ub'):  # pragma: no cover
            self.ub = np.array(None)
        if not hasattr(self, 'simSens'):  # pragma: no cover
            self.simSens = None
        if not hasattr(self, 'sensitivityCalc'):  # pragma: no cover
            self.sensitivityCalc = None
        if not hasattr(self, 'RATE_MIN'):  # pragma: no cover
            self.RATE_MIN = None
        if not hasattr(self, 'RATE_MAX'):  # pragma: no cover
            self.RATE_MAX = None
        if not hasattr(self, 'COST_THRESHOLD'):  # pragma: no cover
            self.COST_THRESHOLD = None
        if not hasattr(self, '_TRUE_PARAMETERS'):  # pragma: no cover
            self._TRUE_PARAMETERS = np.array(None)
        if not hasattr(self, 'sim'):  # pragma: no cover
            self.sim = None
        if not hasattr(self, 'T_MAX'):  # pragma: no cover
            self.T_MAX = None
        if not hasattr(self, 'TIMESTEP'):  # pragma: no cover
            self.TIMESTEP = None
        if not hasattr(self, '_PARAMETER_CONTAINER'):  # pragma: no cover
            self._PARAMETER_CONTAINER = ''
        if not hasattr(self, '_OUTPUT_NAME'):  # pragma: no cover
            self._OUTPUT_NAME = ''
        if not hasattr(self, '_TOLERANCES'):  # pragma: no cover
            self._TOLERANCES = None
        if not hasattr(self, 'NAME'):  # pragma: no cover
            self.NAME = ''
        if not hasattr(self, '_MODEL'):  # pragma: no cover
            self._MODEL = None
        if not hasattr(self, 'WEIGHTS'):  # pragma: no cover
            self.WEIGHTS = None
        if not hasattr(self, '_ANALYTICAL_MODEL'):  # pragma: no cover
            self._ANALYTICAL_MODEL = None
        if not hasattr(self, '_RATE_FUNCTIONS'):  # pragma: no cover
            self._RATE_FUNCTIONS = None

        self.useScaleFactors = False
        self.logTransformParams = [False] * self.n_parameters()  # Are any of the parameter log-transformed
        if 'staircase' in self.NAME:
            # Staircase problems should have bounds turned on
            self.add_parameter_bounds()
            self.add_rate_bounds()
        else:
            # Otherwise they start turned off
            self.parametersBounded = False
            self.ratesBounded = False
        self.plotter = True  # Should the performance metrics be plotted when evaluate() is called
        self.tracker = Tracker()  # Tracks the performance metrics
        self.V_LOW = -120
        self.V_HIGH = 60

        # Set numpy writeable flags
        self._TRUE_PARAMETERS.flags['WRITEABLE'] = False
        self._LOWER_BOUND.flags['WRITEABLE'] = False
        self._UPPER_BOUND.flags['WRITEABLE'] = False

    def load_data(self, dataPath=''):
        """
        Loads output data to use in fitting.

        Parameters
        ----------
        dataPath : string, optional
            An absolute filepath to the .csv data file. The default is '', in which case no file will be loaded.

        Returns
        -------
        None.

        """
        # Load the current/output data from ionBench
        if not dataPath == '':
            tmp = []
            with open(dataPath, newline='') as csvfile:
                reader = csv.reader(csvfile)
                for row in reader:
                    tmp.append(float(row[0]))
            self.DATA = np.array(tmp)

    def sample(self, n=1):  # pragma: no cover
        raise NotImplementedError

    @staticmethod
    def protocol():  # pragma: no cover
        raise NotImplementedError

    def add_parameter_bounds(self):
        """
        Add bounds to the parameters. The bounds will be checked whenever the model is about to be solved, if they are violated then the model will not be solved and a penalty cost will be reported. The model solve count will not be incremented if the bounds are violated but the parameter vector will still be tracked for the other metrics.
        """
        self.lb = np.copy(self._LOWER_BOUND)
        self.ub = np.copy(self._UPPER_BOUND)
        self.parametersBounded = True

    def add_rate_bounds(self):
        """
        Add bounds to the rates. The bounds will be checked whenever the model is about to be solved, if they are violated then the model will not be solved and an infinite cost will be reported. The model solve count will not be incremented if the bounds are violated but the parameter vector will still be tracked for the other metrics.

        The bounds are checked after reversing any log-transforms on parameters (ie on exp(inputted parameters)).
        """
        self.ratesBounded = True

    def clamp_parameters(self, parameters):
        """
        Clamp a parameter vector to be between the current benchmarker bounds. Parameters should be specified in input parameter space. If no bounds are active or the inputted parameter vector obeys the current bounds, the inputted parameter vector is returned.

        Parameters
        ----------
        parameters : list
            Parameter vector in input space to clamp to the current bounds.

        Returns
        -------
        clampedParameters : list
            Parameter vector where the parameters that are out of bounds are clamped to the bounds.

        """
        # Don't edit the inputted parameter vector
        p = np.copy(parameters)
        # If out of bounds
        if not self.in_parameter_bounds(parameters):
            # Find bounds in input space
            lb = self.input_parameter_space(self.lb)
            ub = self.input_parameter_space(self.ub)
            # Clamp to input space bounds
            for i in range(self.n_parameters()):
                if p[i] < lb[i]:
                    p[i] = lb[i]
                elif p[i] > ub[i]:
                    p[i] = ub[i]
        return p

    def log_transform(self, whichParams=None):
        """
        Fit some parameters in a log-transformed space.

        Inputted log-transformed parameters will be set to exp(inputted parameters) before solving the model.

        Parameters
        ----------
        whichParams : list, optional
            Which parameters should be log-transformed, in the form of a list of booleans, the same length as the number of parameters, where True is a parameter to be log-transformed. The default is None, in which case all parameters will be log-transformed.

        Returns
        -------
        None.

        """
        if whichParams is None:  # Log-transform all parameters
            whichParams = [True] * self.n_parameters()
        self.logTransformParams = whichParams

    def input_parameter_space(self, parameters):
        """
        Maps parameters from the original parameter space to the input space. Incorporating any log transforms or scaling factors.

        Parameters
        ----------
        parameters : list
            Parameter vector in the original parameter space.

        Returns
        -------
        parameters : np.ndarray
            Parameter vector mapped to input space.

        """
        # Don't edit the inputted parameter vector
        parameters = np.copy(parameters)
        for i in range(self.n_parameters()):
            # Apply scale factor transforms
            if self.useScaleFactors:
                parameters[i] = parameters[i] / self._TRUE_PARAMETERS[i]
            # Apply log transforms
            if self.logTransformParams[i]:
                parameters[i] = np.log(parameters[i])

        return parameters

    def original_parameter_space(self, parameters):
        """
        Maps parameters from input space to the original parameter space. Removing any log transforms or scaling factors.

        Parameters
        ----------
        parameters : list
            Parameter vector in input space.

        Returns
        -------
        parameters : np.ndarray
            Parameter vector mapped to the original parameter space.

        """
        # Don't edit the inputted parameter vector
        parameters = np.copy(parameters)
        for i in range(self.n_parameters()):  # Note the reverse order compared with input_parameter_space
            # Remove log transforms
            if self.logTransformParams[i]:
                parameters[i] = np.exp(parameters[i])
            # Remove scale factor transform
            if self.useScaleFactors:
                parameters[i] = parameters[i] * self._TRUE_PARAMETERS[i]

        return parameters

    def transform_jacobian(self, parameters):
        """
        Finds the jacobian for the current parameter transform for derivatives calculated in original parameter space to be mapped to the input parameter space.

        Parameters
        ----------
        parameters : array
            Parameter vector at which the jacobian should be calculated. Given in input parameter space.

        Returns
        -------
        derivs : array
            Vector of transform derivatives. To map a derivative that was calculated in original parameter space to one calculated in input parameter space, it should be multiplied element-wise by derivs.

        """
        # Create transform vector of all 1s
        derivs = np.ones(self.n_parameters())
        # Apply transforms to derivs
        for i in range(self.n_parameters()):
            if self.logTransformParams[i]:
                derivs[i] *= np.exp(parameters[i])
            if self.useScaleFactors:
                derivs[i] *= self._TRUE_PARAMETERS[i]

        return derivs

    def in_parameter_bounds(self, parameters, boundedCheck=True):
        """
        Checks if parameters are inside any rectangular parameter bounds. If boundedCheck is True, then it will always return True if benchmarker.parametersBounded = False.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters (in original parameter space) to check against the bounds.
        boundedCheck : bool, optional
            If True, and benchmarker.parametersBounded=False, in_parameter_bounds will always return True. If False, it will ignore the value of bm.parametersBounded and base the returned value only on the parameters and parameter bounds. The default is True.

        Returns
        -------
        bool
            True if parameters are inside the bounds or no bounds are specified, False if the parameters are outside the bounds.

        """
        if not self.parametersBounded and boundedCheck:
            return True
        return self.parameter_penalty(parameters) == 0

    def in_rate_bounds(self, parameters, boundedCheck=True):
        """
        Checks if rates are inside bounds. If boundedCheck is True, then it will always return True if benchmarker.parametersBounded = False.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters (in original parameter space) to check against the bounds.
        boundedCheck : bool, optional
            If True, and benchmarker.parametersBounded=False, in_rate_bounds will always return True. If False, it will ignore the value of bm.parametersBounded and base the returned value only on the parameters and rate bounds. The default is True.

        Returns
        -------
        bool
            True if rates are inside the bounds or no bounds are specified, False if the rates are outside the bounds.

        """
        if not self.ratesBounded and boundedCheck:
            return True
        return self.rate_penalty(parameters) == 0

    def set_max_iter_flag(self, hitMaxIter=True):
        """
        Sets a flag in ionBench to indicate that the optimisation algorithm has hit the maximum number of iterations. This is used to note any optimisation runs that may need repeating, as we try to avoid terminating due to maximum number of iterations. This flag is also set for similar termination criteria like maximum number of function evaluations. The flag is stored in the Tracker.

        Parameters
        ----------
        hitMaxIter : bool, optional
        Set tracker.maxIterFlag to this value. The default is True.

        Returns
        -------
        None.
        """
        self.tracker.maxIterFlag = hitMaxIter

    def n_parameters(self):
        """
        Returns the number of parameters in the model.
        """
        return len(self._TRUE_PARAMETERS)

    def reset(self, fullReset=True):
        """
        Resets the benchmarker. This clears the simulation object and restarts the performance tracker. Optionally, the transforms and bounds can be turned off.

        Parameters
        ----------
        fullReset : bool, optional
            If True, transforms and bounds will be reset (turned off except staircase which will turn bounds on). The default is True.
        """
        # Reset the simulation object
        self.sim.reset()
        # If we have sensitivities enabled, reset that too
        if self.sensitivityCalc:
            self.simSens.reset()
        # Generate a new tracker
        self.tracker = Tracker()
        # Are we resetting the modification too?
        if fullReset:
            self.log_transform([False] * self.n_parameters())
            self.useScaleFactors = False
            # Keep staircase bounds consistent
            if 'staircase' in self.NAME:
                self.add_rate_bounds()
                self.add_parameter_bounds()
            else:
                self.parametersBounded = False
                self.ratesBounded = False
            self.lb = np.copy(self._LOWER_BOUND)
            self.ub = np.copy(self._UPPER_BOUND)
            self.RATE_MIN = 1.67e-5
            self.RATE_MAX = 1e3

    def cost(self, parameters, incrementSolveCounter=True):
        """
        Find the RMSE cost between the model solved using the inputted parameters and the data.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters to solve the model.
        incrementSolveCounter : bool, optional
            If False, it disables the solve counter tracker. This never needs to be set to False by a user. This is only required by the evaluate() method. The default is True.

        Returns
        -------
        cost : float
            The RMSE cost of the parameters.

        """
        testOutput = np.array(
            self.simulate(parameters, np.arange(0, self.T_MAX, self.TIMESTEP),
                          incrementSolveCounter=incrementSolveCounter))
        cost = self.rmse(testOutput, self.DATA)
        return cost

    def signed_error(self, parameters):
        """
        Similar to the cost method, but instead returns the vector of residuals/errors in the model output.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters to solve the model.

        Returns
        -------
        signed_error : numpy array
            The vector of model errors/residuals.

        """
        # Calculate cost for a given set of parameters
        testOutput = np.array(self.simulate(parameters, np.arange(0, self.T_MAX, self.TIMESTEP)))
        return testOutput - self.DATA

    def squared_error(self, parameters):
        """
        Similar to the cost method, but instead returns the vector of squared residuals/errors in the model output.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters to solve the model.

        Returns
        -------
        signed_error : numpy array
            The vector of model squared errors/residuals.

        """
        # Calculate cost for a given set of parameters
        testOutput = np.array(self.simulate(parameters, np.arange(0, self.T_MAX, self.TIMESTEP)))
        return (testOutput - self.DATA) ** 2

    def use_sensitivities(self):
        """
        Turn on sensitivities for a benchmarker. Also sets the parameter self.sensitivityCalc to True. If sensitivities are already enabled (by checking bm.sensitivityCalc), then this method will do nothing.
        """
        # Check sensitivities aren't already enabled
        if not self.sensitivityCalc:
            # Enable sensitivities and compile
            paramNames = [self._PARAMETER_CONTAINER + '.p' + str(i + 1) for i in range(self.n_parameters())]
            self.simSens = myokit.Simulation(self._MODEL, protocol=self.protocol(),
                                             sensitivities=([self._OUTPUT_NAME], paramNames))
            self.simSens.set_tolerance(*self._TOLERANCES)
            self.sensitivityCalc = True

    def grad(self, parameters, incrementSolveCounter=True, inInputSpace=True, returnCost=False, residuals=False):
        """
        Find the gradient of the RMSE cost at the inputted parameters. Gradient is calculated using Myokit sensitivities.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters to find the derivatives about.
        incrementSolveCounter : bool, optional
            If False, it disables the solve counter tracker. This never needs to be set to False by a user. This is only required by the evaluate() method. The default is True.
        inInputSpace : bool, optional
            Specifies whether the inputted parameters are in input space. If True, then the derivative will be transformed into input space as well. The default is True.
        returnCost : bool, optional
            Whether the function should return the cost at parameters, in addition to the gradient. If True, it will return a tuple (cost, grad). If residuals is True, it will return (signed_error, grad). The default is False.
        residuals : bool, optional
            Whether the function should calculate the gradient of the cost, or the jacobian of the residual vector. If True, jacobian of residuals is returned instead of the gradient of the cost function. The default is False.

        Returns
        -------
        grad : array
            The gradient of the RMSE cost, evaluated at the inputted parameters.

        """
        # Undo any transforms
        if inInputSpace:
            parameters = self.original_parameter_space(parameters)
        else:
            parameters = np.copy(parameters)

        # Check model is set up to solve for sensitivities
        if not self.sensitivityCalc:
            warnings.warn(
                "Current benchmarker problem not configured to use derivatives. Will recompile the simulation object with this enabled.")
            self.use_sensitivities()
        # Now to find grad, cost, J (jacobian) and error (residuals)
        # First check if the parameters are out of bounds
        # Abort solving if the parameters are out of bounds
        inParameterBounds = self.in_parameter_bounds(parameters)
        inRateBounds = self.in_rate_bounds(parameters)
        if not inParameterBounds or not inRateBounds:
            # Gradient out of bounds so use penalty function
            # Calculate gradient of penalty function
            parametersMG = [mg.Tensor(p) for p in parameters]
            penalty = self.parameter_penalty(parametersMG) + self.rate_penalty(parametersMG)
            assert penalty.data > 0
            penalty.backward()
            grad = np.zeros(self.n_parameters())
            for i in range(self.n_parameters()):
                grad[i] = parametersMG[i].grad if parametersMG[i].grad is not None else 0
            cost = float(penalty.data)
            error = cost * np.ones(len(self.DATA))
            J = np.zeros((len(self.DATA), self.n_parameters()))
            for i in range(len(self.DATA)):
                J[i,] = grad
            self.tracker.update(parameters, incrementSolveCounter=False, solveType='grad', cost=cost,
                                solveTime=0)
        else:
            # Inside any bounds, so we can get sensitivities
            # Will get sensitivities and find the J (jacobian) and curr (current), then exit the try except else block to calculate grad, cost and error from J and curr
            self.simSens.reset()
            self.set_params(parameters)
            self.set_steady_state(parameters)
            start = time.monotonic()
            try:
                curr, sens = self.solve_with_sensitivities(times=np.arange(0, self.T_MAX, self.TIMESTEP))
                sens = np.array(sens)
            except (myokit.SimulationError, np.linalg.LinAlgError):  # pragma: no cover
                # If the model fails to solve, we will assume the cost is infinite and the jacobian/gradient points back towards good parameters
                end = time.monotonic()
                warnings.warn(
                    'Tried to evaluate gradient but model failed to solve, likely poor choice of parameters. ionBench will try to resolve this by assuming infinite cost and a gradient that points back towards good parameters.')
                self.tracker.update(parameters, incrementSolveCounter=False, solveType='grad',
                                    solveTime=end - start)
                curr = np.array([np.inf] * len(self.DATA))
                # use grad to point back to reasonable parameter space
                # Define jacobian
                J = np.zeros((len(curr), self.n_parameters()))
                tmp = -1 / (self.original_parameter_space(self.sample()) - parameters)
                for i in range(len(curr)):
                    J[i,] = tmp
            else:
                # Here the model successfully solved, so we can calculate the jacobian
                end = time.monotonic()
                # Define the jacobian from the sensitivities
                J = np.zeros((len(curr), self.n_parameters()))
                for i in range(len(curr)):
                    for j in range(self.n_parameters()):
                        J[i, j] = sens[i, 0, j]
            # Calculate the gradient from the jacobian
            grad = []
            error = curr - self.DATA
            cost = self.rmse(curr, self.DATA)
            for i in range(self.n_parameters()):
                if 'moreno' in self.NAME:
                    grad.append(np.dot(error * self.WEIGHTS, J[:, i]) / cost)
                else:
                    grad.append(np.dot(error, J[:, i]) / (len(self.DATA) * cost))
            self.tracker.update(parameters, cost=cost, incrementSolveCounter=incrementSolveCounter,
                                solveType='grad',
                                solveTime=end - start)
        return self.map_derivative(J, grad, parameters, inInputSpace, returnCost, residuals, cost, error)

    def map_derivative(self, J, grad, parameters, inInputSpace, returnCost, residuals, cost, error):
        """
        Map the derivatives (J/jacobian and grad/gradient) onto the correct parameter space and then return the specified outputs. Outputs are described in the docstring for grad.
        """
        # Map derivatives to input space
        if inInputSpace:
            derivs = self.transform_jacobian(self.input_parameter_space(parameters))
            if residuals:
                J *= derivs
            else:
                grad *= derivs
        # Handle return options
        if returnCost:
            if residuals:
                return error, J
            else:
                return cost, grad
        else:
            if residuals:
                return J
            else:
                return grad

    def set_params(self, parameters):
        """
        Set the parameters in the simulation object (both normal and sensitivity if needed). Inputted parameters should be in the original parameter space.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters to solve the model.
        """
        # Update the parameters
        for i in range(self.n_parameters()):
            self.sim.set_constant(self._PARAMETER_CONTAINER + '.p' + str(i + 1), parameters[i])
            if self.sensitivityCalc:
                self.simSens.set_constant(self._PARAMETER_CONTAINER + '.p' + str(i + 1), parameters[i])
        # Workaround for old myokit bug - https://github.com/myokit/myokit/issues/1044 - fixed in 1.36.0
        if 'moreno' in self.NAME:
            self.sim.set_parameters(self.sim.parameters())

    # noinspection PyProtectedMember
    def set_steady_state(self, parameters):
        """
        Sets the model to steady state at -80mV for the specified parameters. Will update the sensitivity simulation if self.sensitivityCalc == True.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters to use for the calculation of the steady state.
        """
        # Initial membrane voltage
        if 'moreno' in self.NAME:
            V = -120
        else:
            V = -80
        At = np.zeros(self.n_parameters())
        if 'hh' in str(type(self._ANALYTICAL_MODEL)):
            # HH does autodiff through myokit
            t = [mg.tensor(a, dtype=np.double) for a in parameters]
            state = self._ANALYTICAL_MODEL._steady_state_function(V, *t)
            s_state = np.zeros((len(parameters), len(state)))
            for j in range(len(state)):
                state[j].backward()
                sens = [float(t[i].grad) if t[i].grad is not None else 0 for i in range(self.n_parameters())]
                [t[i].null_grad() for i in range(self.n_parameters())]
                s_state[:, j] = sens

            state = [float(state[i]) for i in range(len(state))]
            s_state = [list(s_state[i]) for i in range(self.n_parameters())]
        else:
            # MM does autodiff through ionBench reproduction of some myokit code
            assert 'markov' in str(type(self._ANALYTICAL_MODEL))
            n = len(self._ANALYTICAL_MODEL._states)
            f = ionbench.utils.autodiff.get_matrix_function(self._ANALYTICAL_MODEL)

            state = np.zeros((len(parameters), n))
            s_state = np.zeros((len(parameters), n))
            for j in range(n):
                t = [mg.tensor(a, np.double) for a in parameters]
                At, _ = f(*t, V)
                B = At[:-1, -1:]
                A = At[:-1, :-1] - B

                xMG = ionbench.utils.autodiff.linalg_solve(A, -B)
                state = mg.zeros(n)
                for i in range(len(xMG)):
                    state[i] = xMG[i]
                state[-1] = 1 - np.sum(xMG)
                state[j].backward()
                sens = [float(t[i].grad) if t[i].grad is not None else 0 for i in range(self.n_parameters())]
                s_state[:, j] = sens
            state = [float(state[i]) for i in range(len(state))]
            s_state = [list(s_state[i]) for i in range(self.n_parameters())]
        # Check our steady state calculations don't differ from myokit
        try:
            myokitState = self._ANALYTICAL_MODEL.steady_state(V, parameters)
            if np.linalg.norm(np.subtract(myokitState, state)) > 1e-6:  # pragma: no cover
                warnings.warn(f'Myokit and ionBench seems to slightly (norm: {np.linalg.norm(myokitState - state)}) disagree on the steady state. Will use the myokit steady state.')
                state = myokitState
        except myokit.lib.markov.LinearModelError as e:  # pragma: no cover
            if 'eigenvalues' in str(e):
                warnings.warn(
                    'Positive eigenvalues found so steady state is unstable. Will use states defined in problem instead.')
                state = None
            else:
                raise
        except np.linalg.LinAlgError as e:  # pragma: no cover
            if 'infs or NaNs' in str(e):
                # Infs or NaNs in matrix, caused error in myokit steady state
                warnings.warn('Infs or NaNs in matrix so steady state is invalid. Will use states defined in problem instead.')
                state = None
            else:
                raise
        # Now the steady state has been calculated we check it makes sense
        if state is not None:
            if np.any(np.array(state) < 0):
                npstate = np.array(state)
                if np.all(np.abs(npstate[npstate < 0]) < 1e-12):
                    # If the negative values are very close to zero, then we can assume they are zero
                    state = [max(0, s) / np.sum(npstate) for s in npstate]
                else:  # pragma: no cover
                    # Some parameter combinations lead to negative steady state values (not just rounding errors). We don't set the state in this case as a positive steady state doesn't exist.
                    if np.any(At < 0):
                        warnings.warn(
                            'The parameters give negative transition probabilities which leads to an invalid steady state. Will use states defined in the problem instead.')
                        state = None
                    else:
                        print(parameters)
                        print(At)
                        warnings.warn(
                            "Steady state contains significant (>1e-12) negative values. If you see this, please open an issue on the GitHub repo. For now, I will continue with the myokits state (which hopefully doesn't have negative values)")
                        state = myokitState
        # Finally, set the initial state
        if state is not None:
            self.sim.set_state(state)
            if self.sensitivityCalc:
                self.simSens.set_state(state)
                self.simSens._s_state = s_state

    def solve_with_sensitivities(self, times):
        """
        Solve the model with sensitivities to find both the current, curr, and the sensitivities dcurr/dp_i, sens.
        ----------
        times : list
            The times at which to record the current and sensitivities.

        Returns
        -------
        curr : numpy array
            The model outputted current.
        sens : numpy array
            The sensitivities of the outputted current.
        """
        log, e = self.simSens.run(self.T_MAX + 1, log=[self._OUTPUT_NAME], log_times=times)
        return np.array(log[self._OUTPUT_NAME]), e

    def solve_model(self, times, continueOnError=True):
        """
        Solve the model at the inputted times and return the current trace.

        Parameters
        ----------
        times : list or numpy array
            Vector of time points to record model output. Typically, in ms.
        continueOnError : bool, optional
            If continueOnError is True, any errors that occur during solving the model will be ignored and an infinite output will be given. The default is True.

        Returns
        -------
        modelOutput : list
            A vector of model outputs (current trace).

        """
        if continueOnError:
            try:
                if 'staircase' in self.NAME:
                    log = self.sim.run(self.T_MAX + 1, log=[self._OUTPUT_NAME], log_times=times)
                else:
                    log = self.sim.run(self.T_MAX + 1, log_times=times)
                return np.array(log[self._OUTPUT_NAME])
            except (myokit.SimulationError, np.linalg.LinAlgError):  # pragma: no cover
                warnings.warn("Failed to solve model. Will report infinite output in the hope of continuing the run.")
                return np.array([np.inf] * len(times))
        else:
            log = self.sim.run(self.T_MAX + 1, log_times=times)
            return np.array(log[self._OUTPUT_NAME])

    def simulate(self, parameters, times, continueOnError=True, incrementSolveCounter=True):
        """
        Simulate the model for the inputted parameters and return the model output at the specified times.

        Parameters
        ----------
        parameters : list or numpy array
            Vector of parameters to solve the model.
        times : list or numpy array
            Vector of time points to record model output. Typically, in ms.
        continueOnError : bool, optional
            If continueOnError is True, any errors that occur during solving the model will be ignored and an infinite output will be given. The default is True.
        incrementSolveCounter : bool, optional
            If False, it disables the solve counter tracker. This never needs to be set to False by a user. This is only required by the evaluate() method. The default is True.

        Returns
        -------
        modelOutput : list
            A vector of model outputs (current trace).

        """
        # Return the parameters to the original parameter space
        parameters = self.original_parameter_space(parameters)  # Creates a copy of the parameter vector

        # Reset the simulation
        self.sim.reset()

        # Abort solving if the parameters are out of bounds
        penalty = 0
        if self.parametersBounded:
            penalty += self.parameter_penalty(parameters)
        if self.ratesBounded:
            penalty += self.rate_penalty(parameters)
        if penalty > 0:
            self.tracker.update(parameters, cost=penalty, incrementSolveCounter=False)
            return np.add(penalty, self.DATA)

        # Set the parameters in the simulation object
        self.set_params(parameters)
        self.set_steady_state(parameters)

        # Run the simulation and track the performance
        start = time.monotonic()
        out = self.solve_model(times, continueOnError=continueOnError)
        end = time.monotonic()
        self.tracker.update(parameters, cost=self.rmse(out, self.DATA), incrementSolveCounter=incrementSolveCounter,
                            solveTime=end - start)
        return out

    def parameter_penalty(self, parameters):
        """
        Penalty function for out of bound parameters. The penalty for each parameter p that is out of bounds is 1e5+1e5*|p-bound| where bound is either the upper or lower bound for p, whichever was violated.

        Parameters
        ----------
        parameters : numpy array or mygrad.Tensor
            A possibly out-of-bounds parameter vector to penalise.

        Returns
        -------
        penalty : float or mygrad.Tensor
            The penalty to apply for parameter bound violations.
        """
        # Penalty increases linearly with parameters out of bounds
        penalty = 0
        for i in range(self.n_parameters()):
            if parameters[i] < self.lb[i]:
                penalty += 1e5 + 1e5 * np.log(1 + np.abs(parameters[i] - self.lb[i]))
            elif parameters[i] > self.ub[i]:
                penalty += 1e5 + 1e5 * np.log(1 + np.abs(parameters[i] - self.ub[i]))
        return penalty

    def rate_penalty(self, parameters):
        """
        Penalty function for out of bound rates. The penalty for each rate r that is out of bounds is 1e5+1e5*|r-bound| where bound is either the upper or lower bound for r, whichever was violated.

        Parameters
        ----------
        parameters : numpy array or mygrad.Tensor
            A possibly out-of-bounds parameter vector to penalise.

        Returns
        -------
        penalty : float or mygrad.Tensor
            The penalty to apply for rate bound violations.
        """
        penalty = 0
        # Loop though the problem-specific rate functions
        for i, rateFunc in enumerate(self._RATE_FUNCTIONS):
            # Is the rate maximised at high or low voltages
            k = max(rateFunc(parameters, self.V_HIGH), rateFunc(parameters, self.V_LOW))
            if 'moreno' in self.NAME:  # Moreno is naturally out of bounds for some rates so the bounds are wider there
                if k < self.RATE_MIN:
                    penalty += 1e5
                    penalty += 1e5 * np.log(1 + np.abs(k - self.RATE_MIN))
                elif (k > self.RATE_MAX and i not in [3, 5]) or (k > self.RATE_MAX * 10000 and i in [3, 5]):
                    penalty += 1e5
                    penalty += 1e5 * np.log(1 + np.abs(k - self.RATE_MAX))
            else:
                if k < self.RATE_MIN:
                    penalty += 1e5
                    penalty += 1e5 * np.log(1 + np.abs(k - self.RATE_MIN))
                elif k > self.RATE_MAX:
                    penalty += 1e5
                    penalty += 1e5 * np.log(1 + np.abs(k - self.RATE_MAX))
        return penalty

    def rmse(self, c1, c2):
        """
        Returns the RMSE between c1 and c2. This function is overridden in the Moreno problem to do weighted RMSE.

        Parameters
        ----------
        c1 : numpy array
            A list of model outputs, typically current.
        c2 : numpy array
            The data to compare the model output to. Should be the same size as c1.

        Returns
        -------
        rmse : float
            The RMSE between c1 and c2.
        """
        return np.sqrt(np.mean((c1 - c2) ** 2))

    def is_converged(self):
        """
        Returns whether the optimisation has converged yet (satisfied cost threshold or number of unchanged costs).

        Returns
        -------
        bool
            True if the optimisation has converged, False if it has not.
        """
        return self.tracker.cost_threshold(threshold=self.COST_THRESHOLD) or self.tracker.cost_unchanged()

    def evaluate(self):
        """
        Evaluates the best parameters.

        This method reports the performance metrics for the best parameter vector (calling evaluate() does NOT increase the number of cost function evaluations). If benchmarker.plotter = True, then it also plots the performance metrics over the course of the optimisation.
        """
        print('')
        print('========================================')
        print('===      Evaluating Performance      ===')
        print('========================================')
        print('')
        print('Number of cost evaluations:      ' + str(self.tracker.costSolveCount))
        print('Number of grad evaluations:      ' + str(self.tracker.gradSolveCount))
        print('Best cost:                       {0:.6f}'.format(self.tracker.bestCost))
        self.tracker.report_convergence(self.COST_THRESHOLD)
        print('')
        if self.plotter:
            try:
                self.tracker.plot()
                self.sim.reset()
                self.set_params(self.tracker.firstParams)
                self.set_steady_state(self.tracker.firstParams)
                firstOut = self.solve_model(np.arange(0, self.T_MAX, self.TIMESTEP), continueOnError=True)
                self.sim.reset()
                self.set_params(self.tracker.bestParams)
                self.set_steady_state(self.tracker.bestParams)
                lastOut = self.solve_model(np.arange(0, self.T_MAX, self.TIMESTEP), continueOnError=True)
                plt.figure()
                if "moreno" in self.NAME:  # Moreno plots summary statistics not current
                    plt.plot(self.DATA)
                    plt.plot(firstOut)
                    plt.plot(lastOut)
                    plt.ylabel('Summary Statistics')
                else:
                    plt.plot(np.arange(0, self.T_MAX, self.TIMESTEP), self.DATA)
                    plt.plot(np.arange(0, self.T_MAX, self.TIMESTEP), firstOut)
                    plt.plot(np.arange(0, self.T_MAX, self.TIMESTEP), lastOut)
                    plt.ylabel('Current')
                    plt.xlabel('Time (ms)')
                plt.legend(['Data', 'First Parameters', 'Best Parameters'])
                plt.title('Improvement after fitting: ' + self.NAME)
                plt.show()
            except IndexError:
                # If the tracker has no data (e.g. tracker.firstParams is empty), then we can't plot
                pass



"""
Contains the main Modification class and all the paper-specific modification subclasses.
"""


# noinspection PyProtectedMember
class Modification:
    """
    Modifications provide a way to store some benchmarker settings and apply them to a range of benchmarkers. These settings include log-transforms, parameter bounds, and the use of scale factors. A single modification can be generated, and then applied to a range of benchmarkers. Once a Modification object is instantiated, its transforms and bounds can be applied to a benchmarker, bm, by calling modification.apply(bm)
    """

    def __init__(self, name="", logTransform='off', parameterBounds='off', rateBounds='off', scaleFactors='off',
                 customLogTransform=None,
                 customBounds=None, kwargs=None):
        """
        Initialise a Modification object. Once a Modification is built, its setting can be applied to any benchmarker, bm, by calling modification.apply(bm)

        Parameters
        ----------
        name : string, optional
            A name for the modification. Useful for logging. The default is "".
        logTransform : string, optional
            Setting for log transforms. Options are 'off' (default, none of the parameters will be log transformed), 'on' (a problem specific set of parameters will be log transformed), and 'Custom' (Log transformed parameters specified by the customLogTransform input will be log transformed).
        parameterBounds : string, optional
            Setting for parameter bounds. Options are 'off' (default, none of the parameters will be bounds), 'on' (problem specific bounds using the range in the benchmarkers sampler will be used), and 'Custom' (Bounds on parameters specified by the customBounds input will be used).
        rateBounds : string, optional
            Setting for rate bounds on parameter combinations. Options are 'off' (default, rate bounds will not be used), 'on' (rate bounds will be used).
        scaleFactors : string, optional
            Setting for scale factors. Options are 'off' (default, scale factors won't be used), and 'on' (scale factors will be used, with the default parameters representing 1).
        customLogTransform : list, optional
            A list of parameters to log transform. The list should be the same length as the number of parameters in the benchmarker. This makes this option problem specific and will be unlikely to be usable across different benchmarkers. Only required if logTransform = 'Custom'.
        customBounds : list, optional
            A list containing 2 elements, a list of lower bounds and a list of upper bounds. Each sub-list should be the same length as the number of parameters in the benchmarker. This makes this option problem specific and will be unlikely to be usable across different benchmarkers. Only required if parameterBounds = 'Custom'.
        kwargs : dict, optional
            A dictionary of keyword arguments to be passed into an optimisers run function. The default is None.
        """
        if kwargs is None:
            kwargs = {}
        # Store a dictionary of settings
        self.dict = {'log transform': logTransform, 'parameterBounds': parameterBounds, 'rateBounds': rateBounds,
                     'scale factors': scaleFactors}
        self.customLogTransform = customLogTransform
        self.customBounds = customBounds
        if len(name) > 0:
            self.NAME = name
        else:
            self.NAME = None
        self.kwargs = kwargs

    def apply(self, bm):
        """
        Applies the settings in this modification to the inputted benchmarker. This will override any previous settings assigned to the benchmarker.

        Parameters
        ----------
        bm : benchmarker
            A benchmarker problem to apply this modification.
        """
        self.apply_log_transforms(self.dict['log transform'], bm)
        self.apply_parameter_bounds(self.dict['parameterBounds'], bm)
        self.apply_rate_bounds(self.dict['rateBounds'], bm)
        self.apply_scale_factors(self.dict['scale factors'], bm)

    def apply_log_transforms(self, setting, bm):
        """
        Apply the log transform setting to a benchmarker.

        Parameters
        ----------
        setting : string
            Setting for log transforms. Options are 'off' (default, none of the parameters will be log transformed), 'on' (a problem specific set of parameters will be log transformed), and 'Custom' (Log transformed parameters specified by the customLogTransform input will be log transformed, only usable if customLogTransform was set during the creation of this modification, or modification.customLogTransform is set).
        bm : benchmarker
            A benchmarker problem to apply this modification.
        """
        if setting.lower() == 'off':
            # Disable log transforms
            bm.log_transform([False] * bm.n_parameters())
        elif setting.lower() == 'on':
            # Load standard log transforms from benchmarker
            bm.log_transform(bm.STANDARD_LOG_TRANSFORM)
        elif setting.lower() == 'custom':
            bm.log_transform(self.customLogTransform)
        else:
            print(
                "'" + setting + "' is not a valid option for log transforms. Please use either 'off', 'on', or 'Custom'.")

    def apply_parameter_bounds(self, setting, bm):
        """
        Apply the parameter bounds setting to a benchmarker.

        Parameters
        ----------
        setting : string
            Setting for parameter bounds. Options are 'off' (default, none of the parameters will be bounds), 'on' (problem specific bounds using the range in the benchmarkers sampler will be used), and 'Custom' (Bounds on parameters specified by the customBounds input will be used).
        bm : benchmarker
            A benchmarker problem to apply this modification.
        """
        if setting.lower() == 'off':
            bm.parametersBounded = False
        elif setting.lower() == 'on':
            bm.add_parameter_bounds()
        elif setting.lower() == 'custom':
            bm.add_parameter_bounds()
            bm.lb = self.customBounds[0]
            bm.ub = self.customBounds[1]
        else:
            print(
                "'" + setting + "' is not a valid option for bounds. Please use either 'off', 'on', or 'Custom'.")

        # If staircase, override with True
        if 'staircase' in bm.NAME:
            bm.add_parameter_bounds()

    @staticmethod
    def apply_rate_bounds(setting, bm):
        """
        Apply the rate bounds setting to a benchmarker.

        Parameters
        ----------
        setting : string
            Setting for rate bounds. Options are 'off' (default, rate bounds will not be used), and 'on' (rate bounds will be used).
        bm : benchmarker
            A benchmarker problem to apply this modification.
        """
        if setting.lower() == 'off':
            bm.ratesBounded = False
        elif setting.lower() == 'on':
            bm.add_rate_bounds()
        else:
            print(
                "'" + setting + "' is not a valid option for bounds. Please use either 'off', or 'on'.")

        # If staircase, override with True
        if 'staircase' in bm.NAME:
            bm.add_rate_bounds()

    @staticmethod
    def apply_scale_factors(setting, bm):
        """
        Apply the scale factor setting to a benchmarker.

        Parameters
        ----------
        setting : string
            Setting for scale factors. Options are 'off' (default, scale factors won't be used), and 'on' (scale factors will be used, with the default parameters representing 1).
        bm : benchmarker
            A benchmarker problem to apply this modification.
        """
        if setting.lower() == 'on':
            # Set use scale factors
            bm.useScaleFactors = True
        elif setting.lower() == 'off':
            # Disable scale factors
            bm.useScaleFactors = False
        else:
            print("'" + setting + "' is not a valid option for scale factors. Please use either 'on' or 'off'.")


class Abed2013(Modification):
    def __init__(self):
        # Technically upper and lower bounds are done through a custom transformation
        super().__init__(name='Abed2013', logTransform='off', parameterBounds='on', scaleFactors='off')


class Balser1990(Modification):
    def __init__(self):
        super().__init__(name='Balser1990', logTransform='off', parameterBounds='off', scaleFactors='off')


class Beattie2018(Modification):
    def __init__(self):
        super().__init__(name='Beattie2018', logTransform='on', parameterBounds='on', rateBounds='on', scaleFactors='off')


class BenShalom2012(Modification):
    def __init__(self):
        super().__init__(name='BenShalom2012', logTransform='off', parameterBounds='on', scaleFactors='off', kwargs={'crossoverProb': 0.1})


class Bot2012(Modification):
    def __init__(self):
        super().__init__(name='Bot2012', logTransform='off', parameterBounds='on', scaleFactors='on')


class BuenoOrovio2008(Modification):
    def __init__(self):
        super().__init__(name='BuenoOrovio2008', logTransform='off', parameterBounds='on', scaleFactors='off')


class Cabo2022(Modification):
    def __init__(self):
        super().__init__(name='Cabo2022', logTransform='off', parameterBounds='on', scaleFactors='off')


class Cairns2017(Modification):
    def __init__(self):
        super().__init__(name='Cairns2017', logTransform='off', parameterBounds='on', scaleFactors='off')


class Chen2012(Modification):
    def __init__(self):
        super().__init__(name='Chen2012', logTransform='off', parameterBounds='on', scaleFactors='off')


class Clancy1999(Modification):
    def __init__(self):
        super().__init__(name='Clancy1999', logTransform='off', parameterBounds='off', scaleFactors='off')


class Clausen2020(Modification):
    def __init__(self):
        # Bounds should actually change between PSO and Nelder Mead
        super().__init__(name='Clausen2020', logTransform='off', parameterBounds='on', scaleFactors='off')


class Clerx2019(Modification):
    def __init__(self):
        super().__init__(name='Clerx2019', logTransform='on', parameterBounds='on', rateBounds='on', scaleFactors='off')


class Davies2012(Modification):
    def __init__(self):
        super().__init__(name='Davies2012', logTransform='off', parameterBounds='on', scaleFactors='on')


class Dokos2004(Modification):
    def __init__(self):
        # Technically upper and lower bounds are done through a custom transformation
        super().__init__(name='Dokos2004', logTransform='off', parameterBounds='on', scaleFactors='off')


class Du2014(Modification):
    def __init__(self):
        super().__init__(name='Du2014', logTransform='off', parameterBounds='on', scaleFactors='off')


class Groenendaal2015(Modification):
    def __init__(self):
        super().__init__(name='Groenendaal2015', logTransform='off', parameterBounds='on', scaleFactors='on',
                         kwargs={'nGens': 100, 'popSize': 500})


class Guo2010(Modification):
    def __init__(self):
        # Technically upper and lower bounds are done through a custom transformation
        super().__init__(name='Guo2010', logTransform='off', parameterBounds='on', scaleFactors='off')


class Gurkiewicz2007(Modification):
    def __init__(self):
        super().__init__(name='Gurkiewicz2007', logTransform='off', parameterBounds='on', scaleFactors='off')


class JedrzejewskiSzmek2018(Modification):
    def __init__(self):
        super().__init__(name='JedrzejewskiSzmek2018', logTransform='off', parameterBounds='on', scaleFactors='off',
                         kwargs={'popSize': 8})


class Kohjitani2022(Modification):
    def __init__(self):
        super().__init__(name='Kohjitani2022', logTransform='off', parameterBounds='off', scaleFactors='on')


class Liu2011(Modification):
    def __init__(self):
        super().__init__(name='Liu2011', logTransform='off', parameterBounds='on', scaleFactors='off')


class Loewe2016(Modification):
    def __init__(self):
        super().__init__(name='Loewe2016', logTransform='off', parameterBounds='on', scaleFactors='off')


class Maryak1998(Modification):
    def __init__(self):
        super().__init__(name='Maryak1998', logTransform='off', parameterBounds='off', scaleFactors='off')


class Moreno2016(Modification):
    def __init__(self):
        super().__init__(name='Moreno2016', logTransform='off', parameterBounds='on', scaleFactors='off')


class Sachse2003(Modification):
    def __init__(self):
        super().__init__(name='Sachse2003', logTransform='off', parameterBounds='off', scaleFactors='off')


class Seemann2009(Modification):
    def __init__(self):
        super().__init__(name='Seemann2009', logTransform='off', parameterBounds='off', scaleFactors='off')


class Smirnov2020(Modification):
    def __init__(self):
        super().__init__(name='Smirnov2020', logTransform='off', parameterBounds='on', scaleFactors='on')


class Vanier1999(Modification):
    def __init__(self):
        super().__init__(name='Vanier1999', logTransform='off', parameterBounds='on', scaleFactors='on')


class Wilhelms2012a(Modification):
    def __init__(self):
        super().__init__(name='Wilhelms2012a', logTransform='off', parameterBounds='off', scaleFactors='off')


class Wilhelms2012b(Modification):
    def __init__(self):
        super().__init__(name='Wilhelms2012b', logTransform='off', parameterBounds='on', scaleFactors='off')


class Zhou2009(Modification):
    def __init__(self):
        super().__init__(name='Zhou2009', logTransform='off', parameterBounds='off', scaleFactors='off')


class Empty(Modification):
    """
    A modification with default settings.
    """

    def __init__(self, name=""):
        super().__init__(name=name, logTransform='off', parameterBounds='off', scaleFactors='off')



import pints
import numpy as np
import ionbench


def pints_setup(bm, x0, method, maxIter, debug, forceUnbounded=False):
    """
    Set up a Pints model and optimisation controller for a benchmarker.
    Parameters
    ----------
    bm : benchmarker
        A test problem benchmarker.
    x0 : list
        Initial parameter vector from which to start optimisation. If x0=None, a randomly sampled parameter vector is retrieved from bm.sample().
    method : pints.Optimiser
        A Pints optimiser to use. For example, pints.CMAES, pints.PSO, or pints.NelderMead.
    maxIter : int
        The maximum number of iterations to run for.
    debug : bool
        If debug is True, will increase the logging frequency.
    forceUnbounded : bool, optional
        If True, the optimisation will be forced to be unbounded. Default is False.

    Returns
    -------
    model : Model
        A Pints model containing the benchmarker.
    opt : pints.OptimisationController
        An optimisation controller to run the pints optimisation on.
    """
    if x0 is None:
        x0 = bm.sample()
    model = Model(bm)
    if 'moreno' in bm.NAME:  # pragma: no cover
        times = np.arange(len(bm.DATA))
    else:
        times = np.arange(0, bm.T_MAX, bm.TIMESTEP)
    problem = pints.SingleOutputProblem(model, times, model.bm.DATA)
    error = ErrorWithGrad(problem, bm)

    if bm.parametersBounded and not forceUnbounded:
        if bm.ratesBounded:
            boundaries = AdvancedBoundaries(bm)
        else:
            boundaries = pints.RectangularBoundaries(bm.input_parameter_space(bm.lb), bm.input_parameter_space(bm.ub))
        counter = 1
        while not boundaries.check(x0):  # pragma: no cover
            x0 = bm.sample()
            counter += 1
        if counter > 10:  # pragma: no cover
            print(f'Struggled to find parameters in bounds. Required {counter} iterations.')
        opt = pints.OptimisationController(error, x0, method=method, boundaries=boundaries)
    else:
        opt = pints.OptimisationController(error, x0, method=method)
    opt.set_max_iterations(maxIter)
    opt.set_max_unchanged_iterations(threshold=1e-7)
    if debug:
        opt.set_log_interval(iters=1)
    return model, opt


class ErrorWithGrad(pints.RootMeanSquaredError):
    """
    Pints error measure that uses RMSE and additionally defines the gradient of the RMSE.
    """
    def __init__(self, problem, bm):
        super().__init__(problem)
        self.bm = bm

    def evaluateS1(self, x):
        if 'moreno' in self.bm.NAME:  # pragma: no cover
            raise NotImplementedError('Moreno uses weighted RMSE in cost and grad, pints uses unweighted RMSE. No unweighted RMSE grad is available for Moreno.')
        return self.bm.grad(x, returnCost=True)


class Model(pints.ForwardModel):
    """
    A Pints forwards model containing a benchmarker class.
    """

    def __init__(self, bm):
        """
        Initialise a Pints forward model with a benchmarker, linking up the n_parameters() and simulate() methods.

        Parameters
        ----------
        bm : benchmarker
            A test problem benchmarker.

        Returns
        -------
        None.

        """
        self.bm = bm
        super().__init__()

    def n_parameters(self):
        """
        Returns the number of parameters in the model

        Returns
        -------
        n_parameters : int
            Number of parameters in the model.

        """
        return self.bm.n_parameters()

    def simulate(self, parameters, times):
        """
        Simulates the model and returns the model output.

        Parameters
        ----------
        parameters : list
            A list of parameter values, length n_parameters().
        times : list
            A list of times at which to return model output.

        Returns
        -------
        out : list
            Model output, typically a current trace.

        """
        # Reset the simulation
        return self.bm.signed_error(parameters)+self.bm.DATA


class AdvancedBoundaries(pints.Boundaries):
    """
    Pints boundaries to apply to the parameters and the rates.
    """

    def __init__(self, bm):
        """
        Build a Pints boundary object to apply parameter and rate bounds.

        Parameters
        ----------
        bm : benchmarker
            A test problem benchmarker.

        Returns
        -------
        None.

        """
        self.bm = bm

    def n_parameters(self):
        """
        Returns the number of parameters in the model.
        """
        return self.bm.n_parameters()

    def check(self, parameters):
        """
        Check inputted parameters against the parameter and rate bounds.

        Parameters
        ----------
        parameters : list
            Inputted parameter to check, in the input parameter space.

        Returns
        -------
        paramsInsideBounds : bool
            True if the parameters are inside the bound. False if the parameters are outside the bounds.

        """
        parameters = self.bm.original_parameter_space(parameters)
        return self.bm.in_rate_bounds(parameters) and self.bm.in_parameter_bounds(parameters)



import ionbench.problems.staircase
from ionbench.utils.scipy_setup import least_squares


# noinspection PyShadowingNames
def run(bm, x0=None, maxIter=1000, debug=False):
    """
    Runs Trust Region Reflective optimiser from Scipy.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter vector from which to start optimisation. Default is None, in which case a randomly sampled parameter vector is retrieved from bm.sample().
    maxIter : int, optional
        Maximum number of cost function evaluations. The default is 1000.
    debug : bool, optional
        If True, prints out the cost and parameters found by the algorithm. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified by Trust Region Reflective.
    """
    out = least_squares(bm, x0, debug, method='trf', maxIter=maxIter)

    bm.evaluate()
    return out.x


# noinspection PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Wilhelms2012b
    modNum = 2 -> Du2014
    modNum = 3 -> Loewe2016

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Wilhelms2012b.

    """

    if modNum == 1:
        mod = ionbench.modification.Wilhelms2012b()
    elif modNum == 2:
        mod = ionbench.modification.Du2014()
    elif modNum == 3:
        mod = ionbench.modification.Loewe2016()
    else:
        mod = ionbench.modification.Empty(name='trr_scipy')
    return mod



"""
This module contains the PSO algorithm defined by Seemann et al. 2009.
The initial sampling procedure for particle positions and velocities is not described. The ionBench default are used here.
For the acceleration constants, c1 and c2. A linearly decreasing c1 and increasing c2 was recommended so that is used here.
For the inertia weight, w, either exponentially decreasing or randomly sampled in [0.5,1] was recommended. Since the rate of the exponential decrease was not described, random sampling is used here.
There are no bounds on parameters for Seemann et al. 2009 PSO, so the parameters are not transformed to [0,1] and the particle positions are not clamped.
"""
import numpy as np
import ionbench


# noinspection PyShadowingNames
def run(bm, x0=None, n=20, maxIter=1000, debug=False):
    """
    Runs the PSO algorithm defined by Seemann et al. 2009.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter guess around which the population is sampled. If x0=None (the default), then x0 will be sampled with bm.sample().
    n : int, optional
        Number of particles. The default is 20.
    maxIter : int, optional
        Maximum number of iterations. The default is 1000.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """
    if x0 is None:
        x0 = bm.sample()

    cost_func = ionbench.utils.cache.get_cached_cost(bm)

    # noinspection PyShadowingNames
    class Particle(ionbench.utils.particle_optimisers.Particle):
        """
        Seemann et al. 2009 does not use parameter bounds so cannot be mapped to [0,1]. We implement this by overriding the transform and clamp methods.
        """
        def __init__(self):
            super().__init__(bm, cost_func, x0)

        def clamp(self):
            pass

        def transform(self, parameters):
            return parameters

        def untransform(self, parameters):
            return parameters

    particleList = []
    for i in range(n):
        particleList.append(Particle())

    Gcost = [np.inf] * maxIter  # Best cost ever
    Gpos = [None] * maxIter  # Position of best cost ever
    L = 0
    for L in range(maxIter):
        if L > 0:
            Gcost[L] = Gcost[L - 1]
            Gpos[L] = Gpos[L - 1]

        if debug:
            print('-------------')
            print(f'Beginning population: {L}')
            print(f'Best cost so far: {Gcost[L]}')
            print(f'Found at position: {Gpos[L]}')

        for p in particleList:
            p.set_cost()
            if p.currentCost < Gcost[L]:
                Gcost[L] = p.currentCost
                Gpos[L] = np.copy(p.position)

        # Renew velocities
        if L < 1000:  # Dependence of c1 and c2 on maxIter has been removed
            c1 = 2.5 - L / 1000 * 2
        else:  # pragma: no cover
            c1 = 0.5
        c2 = 3 - c1
        w = np.random.uniform(0.5, 1)
        for p in particleList:
            localAcc = c1 * np.random.rand() * (p.bestPosition - p.position)
            globalAcc = c2 * np.random.rand() * (Gpos[L] - p.position)
            p.velocity = w * p.velocity + localAcc + globalAcc
        if debug:
            print("Velocities renewed")
        # Move positions
        for p in particleList:
            p.position += p.velocity

        if debug:
            print("Positions renewed")
            print(f'Finished population {L}')
            print(f'Best cost so far: {Gcost[L]}')
            print(f'Found at position: {Gpos[L]}')

        if bm.is_converged():
            break

    if L >= maxIter-1:
        bm.set_max_iter_flag()

    bm.evaluate()
    return Particle().untransform(Gpos[L])


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Seemann2009

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Seemann2009.

    """
    mod = ionbench.modification.Seemann2009()
    return mod



import os
import pytest
import scipy.stats

import ionbench
import numpy as np
import matplotlib.pyplot as plt
import myokit


class Problem:
    """
    General problem test class for specific problem test classes to inherit from.
    Defines tests for cost, attribute checks, plotting, parameter bounds, rate bounds, tracking, grad, and steady states.
    """

    @pytest.mark.cheap
    def test_cost(self):
        self.bm.reset()
        # Check cost of default params is sufficiently low (0 for loewe)
        assert self.bm.cost(self.bm._TRUE_PARAMETERS) <= self.costBound
        # Check squared error works
        self.bm.squared_error(self.bm.sample())

    def test_evaluate(self):
        # No errors on empty evaluate
        self.bm.plotter = True
        self.bm.reset()
        self.bm.evaluate()
        self.bm.plotter = False

    @pytest.mark.cheap
    def test_hasattr(self):
        self.bm.reset()
        # Check all necessary variables in problems are defined
        assert hasattr(self.bm, "NAME")
        assert hasattr(self.bm, "_MODEL")
        assert hasattr(self.bm, "_OUTPUT_NAME")
        assert hasattr(self.bm, "_PARAMETER_CONTAINER")
        assert hasattr(self.bm, "_TRUE_PARAMETERS")
        assert hasattr(self.bm, "_RATE_FUNCTIONS")
        assert hasattr(self.bm, "STANDARD_LOG_TRANSFORM")
        assert hasattr(self.bm, "DATA")
        assert hasattr(self.bm, "sim")
        assert hasattr(self.bm, "useScaleFactors")
        assert hasattr(self.bm, "parametersBounded")
        assert hasattr(self.bm, "logTransformParams")
        assert hasattr(self.bm, "plotter")
        assert hasattr(self.bm, "tracker")
        assert hasattr(self.bm, "sensitivityCalc")
        assert hasattr(self.bm, "_ANALYTICAL_MODEL")
        assert hasattr(self.bm, "T_MAX")
        assert hasattr(self.bm, "simSens")
        assert hasattr(self.bm, "TIMESTEP")
        assert hasattr(self.bm, "RATE_MIN")
        assert hasattr(self.bm, "RATE_MAX")
        assert hasattr(self.bm, "V_LOW")
        assert hasattr(self.bm, "V_HIGH")
        assert hasattr(self.bm, "_LOWER_BOUND")
        assert hasattr(self.bm, "_UPPER_BOUND")

    @pytest.mark.cheap
    def test_plotter(self, monkeypatch):
        self.bm.reset()
        # Test plotting functionality runs for evaluate
        monkeypatch.setattr(plt, 'show', lambda: None)  # Disables plots from being shown
        self.bm.plotter = True
        self.bm.cost(self.bm.sample())
        self.bm.evaluate()
        self.bm.plotter = False

    @pytest.mark.cheap
    def test_parameter_bounds(self):
        self.bm.reset()
        # Check bounds give infinite cost outside and solve inside bounds
        self.bm.add_parameter_bounds()
        self.bm.lb = 0.99*self.bm._TRUE_PARAMETERS
        # Check bounds give penalty for 1 param out of bounds outside of bounds
        p = np.copy(self.bm._TRUE_PARAMETERS)
        p[0] *= 0.98
        c = self.bm.signed_error(p)[0]
        c2 = self.bm.cost(p)
        assert 1e5 < c < 2e5
        assert 1e5 < c2 < 2e5
        # Check cost is small inside bounds when bounded
        p[0] = self.bm._TRUE_PARAMETERS[0]
        assert self.bm.cost(p) < 1e5
        # Check parametersBounded = False turns off bounds
        self.bm.parametersBounded = False
        p[0] *= 0.98
        assert self.bm.cost(p) < 1e5
        self.bm.parametersBounded = True
        # Check penalty increases with more bound violations
        p = np.copy(self.bm._TRUE_PARAMETERS) * 0.98
        c = self.bm.signed_error(p)[0]
        assert c > 1e5 * self.bm.n_parameters()
        self.bm.parametersBounded = False

    @pytest.mark.cheap
    def test_rate_bounds(self):
        self.bm.reset()
        if 'staircase' in self.bm.NAME:
            assert self.bm.ratesBounded is True
        else:
            assert self.bm.ratesBounded is False
        self.bm.add_rate_bounds()
        assert self.bm.ratesBounded is True
        tmp = self.bm.RATE_MIN
        # Default params inside rate bounds always
        assert self.bm.cost(self.bm._TRUE_PARAMETERS) < 1e5
        # Move rate bounds so default rates are outside
        self.bm.RATE_MIN = self.bm.RATE_MAX * 2
        assert self.bm.cost(self.bm._TRUE_PARAMETERS) > 1e5
        # Turn off rate bounds should allow solving again
        self.bm.ratesBounded = False
        assert self.bm.cost(self.bm._TRUE_PARAMETERS) < 1e5
        self.bm.RATE_MIN = tmp
        # Test rate upper bounds
        self.bm.reset()
        self.bm.add_rate_bounds()
        p = self.bm.sample()
        self.bm.RATE_MAX = self.bm.RATE_MIN
        assert self.bm.cost(p) > 1e5

    @pytest.mark.cheap
    def test_tracker_solve_count(self):
        # Check solve count works with bounds and evaluate
        self.bm.reset()
        # Solve times are empty
        assert len(self.bm.tracker.costTimes) == 0
        assert len(self.bm.tracker.gradTimes) == 0
        # Solve count is 0 after reset
        assert self.bm.tracker.costSolveCount == 0
        # Solve count increments after solving
        self.bm.cost(self.bm._TRUE_PARAMETERS)
        assert self.bm.tracker.costSolveCount == 1
        # but not when out of bounds
        self.bm.add_parameter_bounds()
        self.bm.lb = 0.99*self.bm._TRUE_PARAMETERS
        p = np.copy(self.bm._TRUE_PARAMETERS)
        p[0] = 0.98 * self.bm._TRUE_PARAMETERS[0]
        self.bm.cost(p)
        assert self.bm.tracker.costSolveCount == 1
        self.bm.parametersBounded = False
        # or out of rate bounds
        self.bm.add_rate_bounds()
        tmp = self.bm.RATE_MIN
        self.bm.RATE_MIN = self.bm.RATE_MAX
        self.bm.cost(self.bm._TRUE_PARAMETERS)
        assert self.bm.tracker.costSolveCount == 1
        self.bm.RATE_MIN = tmp
        self.bm.cost(self.bm._TRUE_PARAMETERS*(1+1e-12))
        assert self.bm.tracker.costSolveCount == 2
        self.bm.ratesBounded = False
        # Solve count doesn't increment when evaluating
        self.bm.evaluate()
        assert self.bm.tracker.costSolveCount == 2
        assert self.bm.tracker.gradSolveCount == 0
        # returns to 0 after resetting
        self.bm.reset(fullReset=False)
        assert self.bm.tracker.costSolveCount == 0
        self.bm.parametersBounded = False
        # grad solve counter increments with grad but not normal solve counter
        self.bm.grad(self.bm._TRUE_PARAMETERS)
        assert self.bm.tracker.costSolveCount == 0
        assert self.bm.tracker.gradSolveCount == 1
        self.bm.reset(fullReset=False)
        p1 = self.bm.sample()
        p2 = self.bm.sample()
        c1 = self.bm.cost(p1)
        c2 = self.bm.cost(p2)
        if c1 < c2:
            assert np.abs(c1 - self.bm.tracker.bestCost) < 1e-8
            assert all(p1 == self.bm.tracker.bestParams)
        else:
            assert np.abs(c2 - self.bm.tracker.bestCost) < 1e-8
            assert all(p2 == self.bm.tracker.bestParams)
        self.bm.cost(self.bm._TRUE_PARAMETERS)
        assert all(self.bm._TRUE_PARAMETERS == self.bm.tracker.bestParams)
        self.bm.reset()

    def test_tracker_save_load(self):
        self.bm.reset()
        p = self.bm.sample()
        self.bm.cost(p)
        cost = np.copy(self.bm.tracker.costs[0])
        time = np.copy(self.bm.tracker.costTimes[0])
        self.bm.tracker.save('temporary_test_tracker.pickle')
        self.bm.reset()
        self.bm.tracker.load('temporary_test_tracker.pickle')
        assert cost == self.bm.tracker.costs[0]
        assert time == self.bm.tracker.costTimes[0]
        os.remove('temporary_test_tracker.pickle')

    @pytest.mark.cheap
    def test_convergence(self):
        self.bm.reset()
        p = self.bm.sample()
        c = self.bm.cost(p)
        if c < self.bm.COST_THRESHOLD:
            assert self.bm.is_converged()
            assert self.bm.tracker.cost_threshold(self.bm.COST_THRESHOLD)
        else:
            assert not self.bm.is_converged()
            assert not self.bm.tracker.cost_threshold(self.bm.COST_THRESHOLD)
            self.bm.cost(self.bm._TRUE_PARAMETERS * (1+1e-7))  # Test that non-default parameters give convergence
            assert self.bm.is_converged()
            assert self.bm.tracker.cost_threshold(self.bm.COST_THRESHOLD)
        # Dropping the cost threshold means not converged
        tmp = self.bm.COST_THRESHOLD
        self.bm.COST_THRESHOLD = -1
        assert not self.bm.is_converged()
        assert not self.bm.tracker.cost_threshold(self.bm.COST_THRESHOLD)
        self.bm.COST_THRESHOLD = tmp
        self.bm.reset()
        # Tracker initially not converged
        assert not self.bm.is_converged()
        assert not self.bm.tracker.cost_threshold(self.bm.COST_THRESHOLD)
        assert not self.bm.tracker.cost_unchanged()
        # Cost convergence rather than threshold
        self.bm.reset()
        # Get a point that doesn't satisfy the cost threshold
        p = self.bm.sample()
        while self.bm.cost(p) <= self.bm.COST_THRESHOLD:
            p = self.bm.sample()
        # Shouldn't be converged from this
        assert not self.bm.is_converged()
        assert not self.bm.tracker.cost_threshold(self.bm.COST_THRESHOLD)
        self.bm.reset()
        # Generate 6 random points and order them by cost
        points = [self.bm.sample() for _ in range(6)]
        costs = [self.bm.cost(p) for p in points]
        order = np.argsort(costs)
        points = [points[i] for i in order]
        self.bm.reset()
        # First point is an improvement
        self.bm.cost(points[0])
        # 2 points without improvement aren't enough
        self.bm.cost(points[1])
        self.bm.cost(points[2])
        assert not self.bm.tracker.cost_unchanged(max_unchanged_evals=3)
        # Third point is enough
        self.bm.cost(points[3])
        assert self.bm.tracker.cost_unchanged(max_unchanged_evals=3)
        # Sampling better points resets counter
        self.bm.reset()
        self.bm.cost(points[1])
        self.bm.cost(points[2])
        assert not self.bm.tracker.cost_unchanged(max_unchanged_evals=3)
        c1 = self.bm.cost(points[0])
        assert not self.bm.tracker.cost_unchanged(max_unchanged_evals=3)
        # Three worse points from here needed to converge
        [self.bm.cost(p) for p in points[3:5]]
        assert not self.bm.tracker.cost_unchanged(max_unchanged_evals=3)
        self.bm.cost(points[5])
        assert self.bm.tracker.cost_unchanged(max_unchanged_evals=3)
        # Sampling better points doesn't undo convergence
        self.bm.cost(self.bm._TRUE_PARAMETERS)
        assert self.bm.tracker.cost_unchanged(max_unchanged_evals=3)
        assert self.bm.is_converged()
        self.bm.reset()

    @pytest.mark.cheap
    def test_check_repeated_params(self):
        self.bm.reset()
        self.bm.reset(fullReset=True)
        self.bm.cost(self.bm._TRUE_PARAMETERS)
        self.bm.cost(self.bm.sample())
        assert len(self.bm.tracker.evals) == 2
        self.bm.cost(self.bm._TRUE_PARAMETERS)
        assert len(self.bm.tracker.evals) == 2
        assert self.bm.tracker.check_repeated_param(self.bm._TRUE_PARAMETERS, 'cost')

    @pytest.mark.filterwarnings("ignore:Current:UserWarning")
    def test_grad(self, plotting=False):
        self.bm.reset()
        self.bm.use_sensitivities()
        # Check gradient calculator is accurate
        a = grad_check(bm=self.bm, x0=self.bm.sample(), plotting=plotting)
        assert a < 0.01
        # Same under log transforms
        self.bm.log_transform([True] + [False] * (self.bm.n_parameters() - 1))
        assert grad_check(bm=self.bm, x0=self.bm.sample(), plotting=plotting) < 0.01
        # Same under scale factor and log transforms
        self.bm.useScaleFactors = True
        assert grad_check(bm=self.bm, x0=self.bm.sample(), plotting=plotting) < 0.01
        # Same under scale factor only
        self.bm.log_transform([False] * self.bm.n_parameters())
        assert grad_check(bm=self.bm, x0=self.bm.sample(), plotting=plotting) < 0.01
        # Input space = False
        self.bm.grad(self.bm.sample(), inInputSpace=False)

    @pytest.mark.filterwarnings("ignore:Current:UserWarning")
    def test_grad_bounds(self, plotting=False):
        self.bm.reset()
        mod = ionbench.modification.Modification(parameterBounds='on')
        mod.apply(self.bm)
        self.bm.use_sensitivities()
        x0 = self.bm.sample()
        # Put two parameter outside of bounds
        self.bm.lb[0] = self.bm.ub[0]
        self.bm.lb[2] = self.bm.ub[2]
        # Check gradient calculator is accurate when outside of bounds
        a = grad_check(bm=self.bm, x0=x0, plotting=plotting)
        assert a < 0.01
        # Handles rate bounds as well
        self.bm.add_rate_bounds()
        self.bm.lb = np.copy(self.bm._LOWER_BOUND)
        tmp = self.bm.RATE_MIN
        self.bm.RATE_MIN = self.bm.RATE_MAX
        a = grad_check(bm=self.bm, x0=x0, plotting=plotting)
        assert a < 0.01
        self.bm.useScaleFactors = True
        a = grad_check(bm=self.bm, x0=self.bm.input_parameter_space(x0), plotting=plotting)
        assert a < 0.01
        self.bm.RATE_MIN = tmp

    @pytest.mark.cheap
    def test_steady_state(self):
        self.bm.reset()
        # Test steady state is right for random parameters
        if 'moreno' not in self.bm.NAME:
            self.bm.reset()
            p = self.bm.sample()
            assert self.bm.in_parameter_bounds(p)
            assert self.bm.in_rate_bounds(p)
            out = self.bm.simulate(parameters=p, times=np.arange(0, self.bm.T_MAX, self.bm.TIMESTEP))
            assert np.abs((out[0] - out[1])) < 1e-8
        self.bm.reset()
        p = self.bm.sample()
        self.bm.set_steady_state(p)
        initState = self.bm.sim.state()
        self.bm.set_params(p)
        self.bm.sim.run(1)
        newState = self.bm.sim.state()
        assert all(np.abs(np.subtract(newState, initState)) < 1e-8)
        # Check it also updates simSens
        if not self.bm.sensitivityCalc:
            self.bm.use_sensitivities()
        self.bm.reset()
        p = self.bm.sample()
        self.bm.set_steady_state(p)
        initState = self.bm.simSens.state()
        self.bm.set_params(p)
        self.bm.simSens.run(1)
        newState = self.bm.sim.state()
        assert all(np.abs(np.subtract(newState, initState)) < 1e-8)
        self.bm.reset(fullReset=True)
        # Check sensitivities are steady state as well
        p = self.bm.sample()
        self.bm.set_steady_state(p)
        self.bm.set_params(p)
        initSens = self.bm.simSens._s_state
        if 'loewe' in self.bm.NAME:
            self.bm.simSens.set_tolerance(1e-7, 1e-7)
        self.bm.simSens.run(5)
        newSens = self.bm.simSens._s_state
        # _s_state stays the same
        assert np.all(np.abs(np.subtract(newSens, initSens)) <= 1e-5)
        # _s_state is non-zero
        assert not np.all(np.array(initSens) == 0)
        # sens of current plots will also show the same
        self.bm.reset()
        self.bm.set_params(p)
        self.bm.set_steady_state(p)
        _, e = self.bm.simSens.run(6, log_times=[0, 5])
        e = np.array(e)
        assert all(e[0, 0, :] - e[1, 0, :] < 1e-5)
        if 'loewe' in self.bm.NAME:
            self.bm.simSens.set_tolerance()

    @pytest.mark.cheap
    def test_clamp(self):
        self.bm.reset()
        # Check parameters that start out of bounds get clamped to inside bounds
        p = np.copy(self.bm._TRUE_PARAMETERS)
        self.bm.add_parameter_bounds()
        self.bm.lb = 0.99*self.bm._TRUE_PARAMETERS
        assert self.bm.in_parameter_bounds(p)
        p[0] = 0.5 * self.bm._TRUE_PARAMETERS[0]
        assert not self.bm.in_parameter_bounds(p)
        assert self.bm.in_parameter_bounds(self.bm.clamp_parameters(p))
        # Check that they are at exactly the bounds and not in interior
        assert self.bm.clamp_parameters(p)[0] == self.bm.lb[0]
        # Check that clamping doesn't change the original parameter vector
        assert p[0] != self.bm.clamp_parameters(p)[0]

    @pytest.mark.cheap
    def test_n_parameters(self):
        self.bm.reset()
        n = self.bm.n_parameters()
        assert n > 0
        assert n == len(self.bm._TRUE_PARAMETERS)
        assert n == len(self.bm.STANDARD_LOG_TRANSFORM)
        assert n == len(self.bm._LOWER_BOUND)
        assert n == len(self.bm._UPPER_BOUND)

    @pytest.mark.cheap
    def test_reset(self):
        self.bm.reset(fullReset=True)
        if 'staircase' in self.bm.NAME:
            assert self.bm.parametersBounded is True
            assert self.bm.ratesBounded is True
        else:
            assert self.bm.parametersBounded is False
            assert self.bm.ratesBounded is False
        assert all([self.bm.logTransformParams[i] is False for i in range(self.bm.n_parameters())])
        assert self.bm.useScaleFactors is False
        assert self.bm.tracker.costSolveCount == 0
        mod = ionbench.modification.Clerx2019()
        mod.apply(self.bm)
        self.bm.useScaleFactors = True
        if 'ikur' in self.bm.NAME:
            self.bm.ratesBounded = False
            self.bm.cost(self.bm.input_parameter_space(self.bm._TRUE_PARAMETERS))
            self.bm.ratesBounded = True
        else:
            self.bm.cost(self.bm.input_parameter_space(self.bm._TRUE_PARAMETERS))
        assert self.bm.parametersBounded is True
        assert self.bm.ratesBounded is True
        assert any([self.bm.logTransformParams[i] is True for i in range(self.bm.n_parameters())])
        assert self.bm.useScaleFactors is True
        assert self.bm.tracker.costSolveCount == 1
        self.bm.reset(fullReset=False)
        assert self.bm.parametersBounded is True
        assert self.bm.ratesBounded is True
        assert any([self.bm.logTransformParams[i] is True for i in range(self.bm.n_parameters())])
        assert self.bm.useScaleFactors is True
        assert self.bm.tracker.costSolveCount == 0
        self.bm.reset()  # Full reset is True by default
        if 'staircase' in self.bm.NAME:
            assert self.bm.parametersBounded is True
            assert self.bm.ratesBounded is True
        else:
            assert self.bm.parametersBounded is False
            assert self.bm.ratesBounded is False
        assert all([self.bm.logTransformParams[i] is False for i in range(self.bm.n_parameters())])
        assert self.bm.useScaleFactors is False
        assert self.bm.tracker.costSolveCount == 0
        assert self.bm.tracker.total_solve_time(i=None) == (0, 0)


class Staircase(Problem):
    @pytest.mark.cheap
    def test_sampler(self):
        self.bm.reset()
        # Check sampler is inside the right bounds and doesn't just return default rates, across all transforms
        # Sampler in bounds
        p = self.bm.sample()
        assert self.bm.in_parameter_bounds(p, boundedCheck=False) and self.bm.in_rate_bounds(p, boundedCheck=False)
        # Sampled different to default
        assert sampler_different(self.bm, self.bm._TRUE_PARAMETERS)
        # Same for scale factor parameter space
        self.bm.useScaleFactors = True
        assert sampler_bounds(self.bm, self.bm.input_parameter_space(self.bm._LOWER_BOUND),
                              self.bm.input_parameter_space(self.bm._UPPER_BOUND))
        assert sampler_different(self.bm, np.ones(self.bm.n_parameters()))
        self.bm.useScaleFactors = False
        # Same for log transformed space
        self.bm.log_transform()
        assert sampler_bounds(self.bm, self.bm.input_parameter_space(self.bm._LOWER_BOUND),
                              self.bm.input_parameter_space(self.bm._UPPER_BOUND))
        assert sampler_different(self.bm, np.log(self.bm._TRUE_PARAMETERS))
        self.bm.log_transform([False] * self.bm.n_parameters())
        # Same for scale factor and log transformed space
        self.bm.log_transform()
        self.bm.useScaleFactors = True
        assert sampler_bounds(self.bm, self.bm.input_parameter_space(self.bm._LOWER_BOUND),
                              self.bm.input_parameter_space(self.bm._UPPER_BOUND))
        assert sampler_different(self.bm, np.zeros(self.bm.n_parameters()))
        self.bm.log_transform([False] * self.bm.n_parameters())
        self.bm.useScaleFactors = False
        # n>1
        self.bm.reset()
        assert len(self.bm.sample(5)) == 5

    @pytest.mark.cheap
    def test_transforms(self):
        self.bm.reset()
        # Check transforms map as expected
        # Log transform default rates
        self.bm.log_transform()
        assert param_equal(self.bm.input_parameter_space(self.bm._TRUE_PARAMETERS), np.log(self.bm._TRUE_PARAMETERS))
        # Original space
        assert param_equal(self.bm.original_parameter_space(np.log(self.bm._TRUE_PARAMETERS)), self.bm._TRUE_PARAMETERS)
        self.bm.log_transform([False] * self.bm.n_parameters())
        # Scale factor default rates
        self.bm.useScaleFactors = True
        assert param_equal(self.bm.input_parameter_space(self.bm._TRUE_PARAMETERS), np.ones(self.bm.n_parameters()))
        assert param_equal(self.bm.original_parameter_space(np.ones(self.bm.n_parameters())), self.bm._TRUE_PARAMETERS)
        # Scale factor and log transformed space
        self.bm.log_transform()
        assert param_equal(self.bm.input_parameter_space(self.bm._TRUE_PARAMETERS), np.zeros(self.bm.n_parameters()))
        assert param_equal(self.bm.original_parameter_space(np.zeros(self.bm.n_parameters())), self.bm._TRUE_PARAMETERS)
        self.bm.useScaleFactors = False
        self.bm.log_transform([False] * self.bm.n_parameters())


class Loewe(Problem):
    @pytest.mark.cheap
    def test_sampler(self):
        self.bm.reset()
        # Check sampler is inside the right bounds and doesn't just return default rates, across all transforms
        # Get bounds from a modification
        mod = ionbench.modification.Modification(parameterBounds='on')
        mod.apply(self.bm)
        lb = self.bm.lb
        ub = self.bm.ub
        self.bm.parametersBounded = False
        # Sampler in bounds
        assert sampler_bounds(self.bm, lb, ub)
        # Sampled different to default
        assert sampler_different(self.bm, self.bm._TRUE_PARAMETERS)
        # Same for scale factor parameter space
        self.bm.useScaleFactors = True
        assert sampler_bounds(self.bm, self.bm.input_parameter_space(lb), self.bm.input_parameter_space(ub))
        assert sampler_different(self.bm, np.ones(self.bm.n_parameters()))
        self.bm.useScaleFactors = False
        # Same for log transformed space
        self.bm.log_transform(self.bm.STANDARD_LOG_TRANSFORM)
        assert sampler_bounds(self.bm, self.bm.input_parameter_space(lb), self.bm.input_parameter_space(ub))
        assert sampler_different(self.bm, np.log(self.bm._TRUE_PARAMETERS))
        self.bm.log_transform([False] * self.bm.n_parameters())
        # Same for scale factor and log transformed space
        self.bm.log_transform(self.bm.STANDARD_LOG_TRANSFORM)
        self.bm.useScaleFactors = True
        assert sampler_bounds(self.bm, self.bm.input_parameter_space(lb), self.bm.input_parameter_space(ub))
        assert sampler_different(self.bm, np.zeros(self.bm.n_parameters()))
        self.bm.log_transform([False] * self.bm.n_parameters())
        self.bm.useScaleFactors = False
        # n>1
        self.bm.reset()
        assert len(self.bm.sample(5)) == 5

    @pytest.mark.cheap
    def test_transforms(self):
        self.bm.reset()
        # Check transforms map as expected
        # Log transform default rates
        self.bm.log_transform(self.bm.STANDARD_LOG_TRANSFORM)
        logDefault = [np.log(self.bm._TRUE_PARAMETERS[i]) if self.bm.STANDARD_LOG_TRANSFORM[i] else self.bm._TRUE_PARAMETERS[i]
                      for i in range(self.bm.n_parameters())]
        assert param_equal(self.bm.input_parameter_space(self.bm._TRUE_PARAMETERS), logDefault)
        # Original space
        assert param_equal(self.bm.original_parameter_space(logDefault), self.bm._TRUE_PARAMETERS)
        self.bm.log_transform([False] * self.bm.n_parameters())
        # Scale factor default rates
        self.bm.useScaleFactors = True
        assert param_equal(self.bm.input_parameter_space(self.bm._TRUE_PARAMETERS), np.ones(self.bm.n_parameters()))
        assert param_equal(self.bm.original_parameter_space(np.ones(self.bm.n_parameters())), self.bm._TRUE_PARAMETERS)
        # Scale factor and log transformed space
        self.bm.log_transform(self.bm.STANDARD_LOG_TRANSFORM)
        assert param_equal(self.bm.input_parameter_space(self.bm._TRUE_PARAMETERS),
                           [0.0 if self.bm.STANDARD_LOG_TRANSFORM[i] else 1.0 for i in range(self.bm.n_parameters())])
        assert param_equal(self.bm.original_parameter_space(
            [0.0 if self.bm.STANDARD_LOG_TRANSFORM[i] else 1.0 for i in range(self.bm.n_parameters())]),
            self.bm._TRUE_PARAMETERS)
        self.bm.log_transform([False] * self.bm.n_parameters())
        self.bm.useScaleFactors = False


class TestMoreno(Problem):
    bm = ionbench.problems.moreno2016.INa(sensitivities=True)
    bm.plotter = False
    costBound = 1e-4

    @pytest.mark.cheap
    def test_sampler(self):
        self.bm.reset()
        # Check sampler is inside the right bounds and doesn't just return default rates, across all transforms
        # Sampler in bounds
        assert sampler_bounds(self.bm, 0.75 * self.bm._TRUE_PARAMETERS, 1.25 * self.bm._TRUE_PARAMETERS)
        # Sampled different to default
        assert sampler_different(self.bm, self.bm._TRUE_PARAMETERS)
        # Sampled different to default
        assert sampler_different(self.bm, self.bm._TRUE_PARAMETERS)
        # Same for scale factor parameter space
        self.bm.useScaleFactors = True
        assert sampler_bounds(self.bm, 0.75, 1.25)
        assert sampler_different(self.bm, np.ones(self.bm.n_parameters()))
        self.bm.useScaleFactors = False
        # Same for log transformed space
        self.bm.log_transform()
        assert sampler_bounds(self.bm, np.log(0.75) + np.log(self.bm._TRUE_PARAMETERS),
                              np.log(1.25) + np.log(self.bm._TRUE_PARAMETERS))
        assert sampler_different(self.bm, np.log(self.bm._TRUE_PARAMETERS))
        self.bm.log_transform([False] * self.bm.n_parameters())
        # Same for scale factor and log transformed space
        self.bm.log_transform()
        self.bm.useScaleFactors = True
        assert sampler_bounds(self.bm, np.log(0.75), np.log(1.25))
        assert sampler_different(self.bm, np.zeros(self.bm.n_parameters()))
        self.bm.log_transform([False] * self.bm.n_parameters())
        self.bm.useScaleFactors = False
        # n>1
        self.bm.reset()
        assert len(self.bm.sample(5)) == 5

    @pytest.mark.cheap
    def test_transforms(self):
        self.bm.reset()
        # Check transforms map as expected
        # Log transform default rates
        self.bm.log_transform()
        assert param_equal(self.bm.input_parameter_space(self.bm._TRUE_PARAMETERS), np.log(self.bm._TRUE_PARAMETERS))
        # Original space
        assert param_equal(self.bm.original_parameter_space(np.log(self.bm._TRUE_PARAMETERS)), self.bm._TRUE_PARAMETERS)
        self.bm.log_transform([False] * self.bm.n_parameters())
        # Scale factor default rates
        self.bm.useScaleFactors = True
        assert param_equal(self.bm.input_parameter_space(self.bm._TRUE_PARAMETERS), np.ones(self.bm.n_parameters()))
        assert param_equal(self.bm.original_parameter_space(np.ones(self.bm.n_parameters())), self.bm._TRUE_PARAMETERS)
        # Scale factor and log transformed space
        self.bm.log_transform()
        assert param_equal(self.bm.input_parameter_space(self.bm._TRUE_PARAMETERS), np.zeros(self.bm.n_parameters()))
        assert param_equal(self.bm.original_parameter_space(np.zeros(self.bm.n_parameters())), self.bm._TRUE_PARAMETERS)
        self.bm.log_transform([False] * self.bm.n_parameters())
        self.bm.useScaleFactors = False


class TestHH(Staircase):
    bm = ionbench.problems.staircase.HH(sensitivities=True)
    bm.plotter = False
    costBound = 0.04  # Accounts for noise


class TestMM(Staircase):
    bm = ionbench.problems.staircase.MM(sensitivities=True)
    bm.plotter = False
    costBound = 0.02  # Accounts for noise


class TestLoeweIKr(Loewe):
    bm = ionbench.problems.loewe2016.IKr(sensitivities=True)
    bm.plotter = False
    costBound = 1e-16


class TestLoeweIKur(Loewe):
    bm = ionbench.problems.loewe2016.IKur(sensitivities=True)
    bm.plotter = False
    costBound = 1e-16


class TestTest(Problem):
    bm = ionbench.problems.test.Test()
    bm.plotter = False
    costBound = 1e-16

    def test_grad_bounds(self, plotting=False):
        pass

    def test_steady_state(self):
        pass

    def test_rate_bounds(self):
        pass


def sampler_bounds(bm, lb, ub):
    """
    Test function for checking that sampled parameters lie within the expected bounds
    """
    p = bm.sample()
    return all(np.logical_and(p >= lb, p <= ub))


def sampler_different(bm, default):
    """
    Test function for checking that sampled parameters are different to the default parameters
    """
    p = bm.sample()
    return not param_equal(p, default)


def param_equal(p1, p2):
    """
    Test function to check if two parameters are equal
    """
    return all(np.abs(p1 - p2) < 1e-10)


def grad_check(bm, x0, plotting=False):
    """
    Test function for checking gradient matches perturbing the cost function
    """
    nPoints = 50
    paramVec = np.linspace(0.999 * x0[0], 1.001 * x0[0], nPoints)
    if 'staircase' in bm.NAME:
        bm.sim.set_tolerance(bm._TOLERANCES[0]/100, bm._TOLERANCES[1]/100)  # Use tighter tolerances here to avoid solver noise
    costs = np.zeros(nPoints)
    for i in range(nPoints):
        p = np.copy(x0)
        p[0] = paramVec[i]
        costs[i] = bm.cost(p)
    if 'staircase' in bm.NAME:
        bm.sim.set_tolerance(*bm._TOLERANCES)
    grad = bm.grad(x0)

    centreCost = bm.cost(x0)
    if plotting:
        plt.plot(paramVec, costs)
        plt.plot(paramVec, centreCost + grad[0] * (paramVec - x0[0]))
        plt.legend(['ODE cost', 'Grad'])
        plt.show()
    actualGrad, _, _, _, _ = scipy.stats.linregress(paramVec, costs)
    return (actualGrad - grad[0])**2 / actualGrad



"""
Describes the genetic algorithm from Cairns et al. 2017.
This algorithm uses a crossover method which is based on the binary representation of the parameters. It does not work (is not defined for) negative parameters and struggles to identify parameters under 1e-6.
"""
import numpy as np
import ionbench
import ionbench.utils.population_optimisers as pop_opt
import copy


# noinspection PyShadowingNames
def run(bm, x0=None, nGens=10, popSize=2500, tournamentSize=5, debug=False):
    """
    Runs the genetic algorithm from Cairns et al. 2017.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter guess. Population is generated by randomly perturbing this initial guess +-50%, clamped to bounds if necessary. If x0=None (the default), then the population will be sampled using bm.sample().
    nGens : int, optional
        The number of generations to run the optimisation algorithm for. The default is 10.
    popSize : int, optional
        The size of the population in each generation. The default is 2500.
    tournamentSize : int, optional
        The size of the tournament to run during tournament selection. The default is 5.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """
    cost_func = ionbench.utils.cache.get_cached_cost(bm)

    if not bm.parametersBounded:
        raise RuntimeError('Cairns et al 2017 GA optimiser requires bounds.')

    boundsRange = bm.input_parameter_space(bm.ub) - bm.input_parameter_space(bm.lb)

    pop = pop_opt.get_pop(bm, x0, popSize, cost_func)

    gen = None
    for gen in range(nGens):
        elites = pop_opt.get_elites(pop, 1)
        if debug:
            print("------------")
            print(f'Gen {gen}, Best cost: {elites[0].cost}')

        # Tournament selection
        parents = []
        # Generate 2*popSize number of parents
        for j in range(popSize * 2):
            # Tournament indices
            tour = np.random.choice(np.arange(popSize), tournamentSize, replace=False)
            # Find the best individual in the tournament
            bestInd = pop[tour[0]]
            for i in tour[1:]:
                if pop[i].cost < bestInd.cost:
                    bestInd = pop[i]
            # Save best individual as parent
            parents.append(copy.deepcopy(bestInd))

        # Crossover
        newPop = []
        for j in range(popSize):
            # Get parents 1 and 2
            par1 = parents[2 * j]
            par2 = parents[2 * j + 1]
            # Construct child
            child = pop_opt.Individual(bm, bm.sample(), cost_func)
            for i in range(bm.n_parameters()):
                # Get parameter i from parents 1 and 2
                num1 = par1.x[i]
                num2 = par2.x[i]
                # Want to convert to integer and maintain 6 decimal places of accuracy
                num1 = int(np.round(num1 * 1e6))
                num2 = int(np.round(num2 * 1e6))
                # Then map to binary string
                # Find length of binary strings
                if num1 == num2:
                    # Don't do crossover as not different, also avoids avoids issue with log2(0)
                    numNew = num1/1e6
                else:
                    binStrLen = int(np.log2(max(num1, num2))) + 1
                    bin1 = format(num1, '0' + str(binStrLen) + 'b')
                    bin2 = format(num2, '0' + str(binStrLen) + 'b')
                    # Random position between 0 and binStrLen (inclusive)
                    randPos = np.random.randint(0, binStrLen + 1)
                    # Crossover in binary
                    binNew = bin1[:randPos] + bin2[randPos:]
                    # Convert back to decimal float
                    numNew = int(binNew, 2) / 1e6
                child.x[i] = numNew
            newPop.append(child)

        # Mutation
        for j in range(popSize):
            for i in range(bm.n_parameters()):
                # Change of mutation is 0.1%
                if np.random.rand() < 0.001:
                    perturb = np.random.normal(0.05 * boundsRange[i], 0.05 * boundsRange[i] / 3)
                    if np.random.rand() < 0.5:
                        newPop[j].x[i] += perturb
                    else:
                        newPop[j].x[i] -= perturb
        pop = newPop

        if debug:
            print(f'Finishing gen {gen}')

        # Find costs
        pop = pop_opt.find_pop_costs(pop)

        # Elitism
        pop = pop_opt.set_elites(pop, elites)

        if bm.is_converged():
            break

    if gen >= nGens - 1:
        bm.set_max_iter_flag()

    elites = pop_opt.get_elites(pop, 1)
    bm.evaluate()
    return elites[0].x


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Cairns2017

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Cairns2017.

    """
    mod = ionbench.modification.Cairns2017()
    return mod



# Extra tests to provide full coverage over ionBench where other tests do not reach
from importlib import import_module
import numpy as np
import ionbench
import pytest


# Problems
def test_problems_coverage():
    bm = ionbench.problems.loewe2016.IKr()
    bm.solve_model(np.arange(0, bm.T_MAX, bm.TIMESTEP), continueOnError=False)
    bm.grad(bm.sample())  # Triggers use_sensitivities() through grad
    bm = ionbench.problems.moreno2016.INa()
    bm.solve_model(np.arange(0, bm.T_MAX, bm.TIMESTEP), continueOnError=False)
    bm = ionbench.problems.test.Test()
    assert len(bm.sample(5)) == 5


# Optimisers
def test_scipy_optimisers():
    """
    Two scipy optimisers aren't fully covered by their modifications, so we add tests here.
    """
    bm = ionbench.problems.test.Test()
    mod = ionbench.modification.Modification(parameterBounds='off', rateBounds='on')
    mod.apply(bm)
    ionbench.optimisers.scipy_optimisers.slsqp_scipy.run(bm, maxIter=10, debug=False)
    bm.reset(False)
    mod = ionbench.modification.Modification(parameterBounds='on')
    mod.apply(bm)
    ionbench.optimisers.scipy_optimisers.powell_scipy.run(bm, maxIter=10, debug=False)


def test_convergence_terminates():
    """
    Test that, in all external optimisers, if the cost threshold is reached, the optimisation terminates.
    """
    bm = ionbench.problems.test.Test()
    for opt in ionbench.OPT_EXT:
        module = import_module(opt)
        module.run(bm)
        assert bm.is_converged()


def test_abort_incompatible_modifications():
    """
    If the modification is incompatible with the optimiser, the optimiser should abort.
    """
    bm = ionbench.problems.test.Test()
    with pytest.raises(RuntimeError):
        ionbench.optimisers.external_optimisers.PSO_Loewe2016.run(bm)
    with pytest.raises(RuntimeError):
        ionbench.optimisers.external_optimisers.PSOTRR_Loewe2016.run(bm)
    with pytest.raises(RuntimeError):
        ionbench.optimisers.external_optimisers.hybridPSOTRR_Loewe2016.run(bm)
    with pytest.raises(RuntimeError):
        ionbench.optimisers.external_optimisers.hybridPSOTRRTRR_Loewe2016.run(bm)


def test_loewe2016():
    """
    Checks Loewe PSO warns about poor choice of phi1 and phi2.
    """
    bm = ionbench.problems.test.Test()
    bm.add_parameter_bounds()
    with pytest.warns(UserWarning):
        ionbench.optimisers.external_optimisers.PSO_Loewe2016.run(bm, phi1=1, phi2=1, maxIter=5)
    with pytest.warns(UserWarning):
        ionbench.optimisers.external_optimisers.PSOTRR_Loewe2016.run(bm, phi1=1, phi2=1, maxIter=5)
    with pytest.warns(UserWarning):
        ionbench.optimisers.external_optimisers.hybridPSOTRR_Loewe2016.run(bm, phi1=1, phi2=1, maxIter=5)
    with pytest.warns(UserWarning):
        ionbench.optimisers.external_optimisers.hybridPSOTRRTRR_Loewe2016.run(bm, phi1=1, phi2=1, maxIter=5)


def test_gurkiewicz2007():
    """
    Checks GA Gurkiewicz warns about using a population size that is too small.
    """
    bm = ionbench.problems.test.Test()
    with pytest.warns(UserWarning):
        ionbench.optimisers.external_optimisers.GA_Gurkiewicz2007a.run(bm, popSize=1, nGens=5)
    with pytest.warns(UserWarning):
        ionbench.optimisers.external_optimisers.GA_Gurkiewicz2007b.run(bm, popSize=1, nGens=5)


def test_cairns2017():
    """
    Checks GA Cairns returns an error when not using bounded parameters.
    """
    bm = ionbench.problems.test.Test()
    with pytest.raises(RuntimeError):
        ionbench.optimisers.external_optimisers.GA_Cairns2017.run(bm)


def test_SA_Vanier_long():
    bm = ionbench.problems.test.Test()
    mod = ionbench.optimisers.external_optimisers.SA_Vanier1999.get_modification()
    mod.apply(bm)
    bm.COST_THRESHOLD = 0
    ionbench.optimisers.external_optimisers.SA_Vanier1999.run(bm, maxIter=100000, debug=True)


def test_PSO_Seemann_long():
    bm = ionbench.problems.test.Test()
    mod = ionbench.optimisers.external_optimisers.pso_Seemann2009.get_modification()
    mod.apply(bm)
    bm.COST_THRESHOLD = 0
    ionbench.optimisers.external_optimisers.pso_Seemann2009.run(bm, maxIter=5000)


def test_PSO_Cabo_long():
    bm = ionbench.problems.test.Test()
    mod = ionbench.optimisers.external_optimisers.pso_Cabo2022.get_modification()
    mod.apply(bm)
    bm.COST_THRESHOLD = 0
    ionbench.optimisers.external_optimisers.pso_Cabo2022.run(bm, maxIter=5000)


def test_Kohjitani_long():
    bm = ionbench.problems.test.Test()
    mod = ionbench.optimisers.external_optimisers.patternSearch_Kohjitani2022.get_modification()
    mod.apply(bm)
    bm.COST_THRESHOLD = 0
    ionbench.optimisers.external_optimisers.patternSearch_Kohjitani2022.run(bm, maxIter=5000)
    bm.reset(False)
    ionbench.optimisers.external_optimisers.patternSearch_Kohjitani2022.run(bm, maxIter=5)
    bm.reset(False)
    ionbench.optimisers.external_optimisers.patternSearch_Kohjitani2022.run(bm, maxfev=5)


def test_NMPSO_Liu_long():
    bm = ionbench.problems.test.Test()
    mod = ionbench.optimisers.external_optimisers.NMPSO_Liu2011.get_modification()
    mod.apply(bm)
    bm.COST_THRESHOLD = 0
    ionbench.optimisers.external_optimisers.NMPSO_Liu2011.run(bm, maxIter=1000, eps=0, debug=True)


def test_Zhou_long():
    bm = ionbench.problems.test.Test()
    mod = ionbench.optimisers.external_optimisers.DE_Zhou2009.get_modification()
    mod.apply(bm)
    bm.COST_THRESHOLD = 0
    ionbench.optimisers.external_optimisers.DE_Zhou2009.run(bm, x0=bm.sample(), debug=True)


def test_hybrid_Loewe_long():
    bm = ionbench.problems.test.Test()
    mod = ionbench.optimisers.external_optimisers.hybridPSOTRR_Loewe2016.get_modification()
    mod.apply(bm)
    bm.COST_THRESHOLD = 0
    ionbench.optimisers.external_optimisers.hybridPSOTRR_Loewe2016.run(bm, x0=bm.sample(), debug=True)
    bm.reset(False)
    ionbench.optimisers.external_optimisers.hybridPSOTRRTRR_Loewe2016.run(bm, x0=bm.sample(), debug=True)


def test_dokos_long():
    bm = ionbench.problems.test.Test()
    # Add noise to get good weight updates
    bm.DATA += np.random.normal(0, 0.003, len(bm.DATA))
    mod = ionbench.optimisers.external_optimisers.curvilinearGD_Dokos2004.get_modification()
    mod.apply(bm)
    bm.COST_THRESHOLD = -1
    ionbench.optimisers.external_optimisers.curvilinearGD_Dokos2004.run(bm, costThreshold=-1, maxIter=250, maxInnerIter=4, debug=True)



"""
This module provides the perturbed PSO from Chen et al. 2012.
The implementation of the algorithm is clear, so we don't need to make any assumptions.
The position sampling does not allow the use of an initial guess x0, so we do not use it with ionBench. Instead, using the ionBench default.
The initial velocity sampling can be used, however.
If the groups are not specified, the case for the presented results, then each parameter is in its own group.
"""
import numpy as np
import ionbench
import itertools


# noinspection PyShadowingNames
def run(bm, x0=None, groups=None, n=6, c1=1.4, c2=1.4, qmax=4, maxIter=1000, w=0.6, debug=False):
    """
    Runs the perturbed particle swarm optimisation algorithm from Chen et al. 2012. If the benchmarker is bounded, the solver will search in the interval [lb,ub], otherwise the solver will search in the interval [0,2*default]

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter guess. Population is generated by randomly perturbing this initial guess +-50%, then applying the appropriate bounds. If x0=None (the default), then the population will be sampled uniformly between the bounds.
    groups : list, optional
        Groupings of parameters to use in the algorithm. Default is None, in which case every parameter will be in its own group.
    n : int, optional
        Number of particles. The default is 6.
    c1 : float, optional
        Scale of the acceleration towards a particle's best positions. The default is 1.4.
    c2 : float, optional
        Scale of the acceleration towards the best point seen across all particles. The default is 1.4.
    qmax : int, optional
        Maximum number of iterations without improvement before perturbations. The default is 4.
    maxIter : int, optional
        Maximum number of iterations. The default is 1000.
    w : float, optional
        Initial value for the inertia of the particles. The default is 0.6.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """

    if not bm.parametersBounded:  # pragma: no cover
        raise RuntimeError('This optimiser requires bounds.')

    cost_func = ionbench.utils.cache.get_cached_cost(bm)

    if x0 is None:
        x0 = bm.sample()

    # noinspection PyShadowingNames
    class Particle(ionbench.utils.particle_optimisers.Particle):
        def __init__(self):
            super().__init__(bm, cost_func, x0)

        def set_velocity(self):
            self.velocity = 0.1 * np.random.rand(bm.n_parameters())

    if groups is None:
        groups = [[i] for i in range(bm.n_parameters())]

    q = 0  # Number of generations without improvement
    # Generate patterns
    patterns = []  # All combinations of groups
    for i in range(len(groups)):
        for j in itertools.combinations(range(len(groups)), i + 1):
            newPattern = []
            for k in j:
                newPattern += groups[k]
            patterns += [newPattern]
    N = len(patterns)  # Number of patterns

    particleList = []
    for i in range(n):
        particleList.append(Particle())

    Gcost = [np.inf] * maxIter  # Best cost ever
    Gpos = [None] * maxIter  # Position of best cost ever
    alpha = [None] * maxIter
    L = None
    for L in range(maxIter):
        if L > 0:
            Gcost[L] = Gcost[L - 1]
            Gpos[L] = Gpos[L - 1]

        if debug:
            print('-------------')
            print(f'Beginning population: {L}')
            print(f'Best cost so far: {Gcost[L]}')
            print(f'Found at position: {Gpos[L]}')

        foundImprovement = False
        for p in particleList:
            p.set_cost()
            if p.currentCost < Gcost[L]:
                Gcost[L] = p.currentCost
                Gpos[L] = np.copy(p.position)
                foundImprovement = True

        if foundImprovement:
            q = 0
            if debug:
                print("Found improvement")
                print(f'Best cost is now:  {Gcost[L]}')
                print(f'Found at position: {Gpos[L]}')
        else:
            q += 1
            if debug:
                print("Didn't find an improvement")
                print(f'Current value of q: {q}')

        if bm.is_converged():
            break

        if q > 5 * qmax:  # pragma: no cover
            # Abort
            print("Too many iterations without improvement")
            print(f'Final cost of {Gcost[L]} found at:')
            print(Gpos[L])
            break

        if q >= qmax:
            if debug:
                print("q exceeds qmax so perturbing")
                print(f'q: {q}, qmax: {qmax}')
            # steps 5.1 and 5.2
            newParticleList = []
            bestNewCost = np.inf
            bestNewPosition = None
            for i in range(N):
                newParticle = Particle()
                newParticle.position = np.copy(Gpos[L])
                for j in patterns[i]:
                    newParticle.position[j] *= 1 + (np.random.rand() - 0.5) / 40
                newParticle.clamp()
                newParticle.set_cost()
                if newParticle.currentCost < bestNewCost:
                    bestNewCost = newParticle.currentCost
                    bestNewPosition = np.copy(newParticle.position)
                newParticleList.append(newParticle)
            if debug:
                print("Perturbed particles")
                print(f'Best new cost is: {bestNewCost} found at {bestNewPosition}')
            if bestNewCost <= Gcost[L]:
                if debug:
                    print("Cost improved by perturbing")
                    print(f'Original best cost: {Gcost[L]}, new best cost: {bestNewCost}')
                q = 0
                Gcost[L] = bestNewCost
                Gpos[L] = np.copy(bestNewPosition)
                # Find worst cost and position
                worstCost = -np.inf
                worstPosition = None
                for p in particleList:
                    if p.currentCost > worstCost:
                        worstCost = p.currentCost
                        worstPosition = p.position
                # Found worst cost and position, now to find that particle and set it to the best position
                if debug:
                    print("Adjusting worst particle")
                for p in particleList:
                    if p.currentCost == worstCost and all(p.position == worstPosition):
                        if debug:
                            print("Worst particle found")
                            print(f'Worst particle cost: {worstCost}, worst particle position: {worstPosition}')
                        p.bestCost = bestNewCost
                        p.bestPosition = np.copy(bestNewPosition)
                        if debug:
                            print(f'New best cost for worst particle: {p.bestCost}, position: {p.position}')
            else:
                if debug:
                    print("Cost wasn't improved by perturbing")

        # Step 6
        alpha[L] = 1 / n * np.sum(np.abs([p.currentCost - Gcost[L] for p in particleList]))
        if L >= 2:
            # Adapt inertia weight w from (17)
            w = np.exp(-alpha[L - 1] / alpha[L - 2])
        if debug:
            print("Updating inertia")
            print(f'w: {w}')
            print(f'alpha: {alpha[L]}')

        # Renew velocities according to (16)
        for p in particleList:
            localAcc = c1 * np.random.rand() * (p.bestPosition - p.position)
            globalAcc = c2 * np.random.rand() * (Gpos[L] - p.position)
            p.velocity = w * p.velocity + localAcc + globalAcc
        if debug:
            print("Velocities renewed")
        # Move positions according to (15) while maintaining [0,1] bounds
        for p in particleList:
            p.position += p.velocity
            p.clamp()

        if debug:
            print("Positions renewed")
            print(f'Finished population {L}')
            print(f'Best cost so far: {Gcost[L]}')
            print(f'Found at position: {Gpos[L]}')

    if L >= maxIter-1:
        bm.set_max_iter_flag()

    bm.evaluate()
    return Particle().untransform(Gpos[L])


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Chen2012

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Chen2012.

    """
    mod = ionbench.modification.Chen2012()
    return mod
